from __future__ import annotations
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    "snells_law_scalar",
    "snells_law",
]


def snells_law_scalar(
    cos_incidence: float | na.AbstractScalar,
    index_refraction: float | na.AbstractScalar,
    index_refraction_new: float | na.AbstractScalar,
) -> na.AbstractScalar:
    """
    A simple form of Snell's law which computes the cosine of the angle
    between the propagation direction inside the new medium and the interface
    normal.

    Parameters
    ----------
    cos_incidence
        The cosine of the incidence angle (the angle between the propagation
        direction and the interface normal.)
    index_refraction
        The index of refraction in the current medium.
    index_refraction_new
        The index of refraction in the new medium.
    """
    if na.unit(cos_incidence) is not None:
        cos_incidence = cos_incidence.to(u.dimensionless_unscaled).value
    sin_incidence = np.emath.sqrt(1 - np.square(cos_incidence))
    sin_transmitted = index_refraction * sin_incidence / index_refraction_new
    cos_transmitted = np.emath.sqrt(1 - np.square(sin_transmitted))
    return cos_transmitted


def snells_law(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    index_refraction: float | na.AbstractScalar,
    index_refraction_new: float | na.AbstractScalar,
    normal: None | na.AbstractCartesian3dVectorArray,
    is_mirror: bool | na.AbstractScalar = False,
    diffraction_order: int = 0,
    spacing_rulings: None | u.Quantity | na.AbstractScalar = None,
    normal_rulings: None | na.AbstractCartesian3dVectorArray = None,
) -> na.Cartesian3dVectorArray:
    r"""
    A `vector form of Snell's law <https://en.wikipedia.org/wiki/Snell%27s_law#Vector_form>`_,
    which has been modified to include the effects of diffraction from a periodic ruling pattern.

    Parameters
    ----------
    wavelength
        The wavelength of the incoming light
    direction
        The propagation direction of the incoming light
    index_refraction
        The index of refraction in the current propagation medium, :math:`n_1`.
    index_refraction_new
        The index of refraction in the new propagation medium, :math:`n_2`.
    normal
        The unit vector perpendicular to the interface between :math:`n_1`
        and :math:`n_2`.
    is_mirror
        A boolean flag controlling whether to compute the direction of the
        reflected ray or the transmitted ray.
        The default is to compute the transmitted ray.
    diffraction_order
        The diffraction order, :math:`m` of the reflected/transmitted rays.
        If nonzero, ``spacing_rulings`` and ``normal_rulings`` must be specified.
    spacing_rulings
        The distance between two grooves in the ruling pattern.
        If ``diffraction_order`` is zero, then this argument has no effect.
    normal_rulings
        The vector perpendicular to the plane of the grooves.

    Examples
    --------
    Plot the reflected and transmitted rays from a specified input ray.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the propagation direction of the
        # incident ray
        angle = 45 * u.deg
        direction = na.Cartesian3dVectorArray(
            x=np.sin(angle),
            y=0,
            z=-np.cos(angle),
        )

        # Define the keyword arguments that are common
        # to both the reflected and transmitted ray
        kwargs = dict(
            wavelength=350 * u.nm,
            direction=direction,
            index_refraction=1,
            normal=na.Cartesian3dVectorArray(0, 0, 1),
        )

        # Calculate the propagation direction of the
        # reflected ray
        direction_reflected = optika.materials.snells_law(
            is_mirror=True,
            index_refraction_new=1,
            **kwargs,
        )

        # Calculate the propagation direction of the
        # transmitted ray
        direction_transmitted = optika.materials.snells_law(
            is_mirror=False,
            index_refraction_new=2,
            **kwargs,
        )

        # Plot the incident, reflected, and transmitted rays
        fig, ax = plt.subplots()
        na.plt.plot(
            na.stack([-direction, 0], axis="plot"),
            axis="plot",
            components=("x", "z"),
            label="incident",
        );
        na.plt.plot(
            na.stack([0, direction_transmitted], axis="plot"),
            axis="plot",
            components=("x", "z"),
            label="transmitted",
        );
        na.plt.plot(
            na.stack([0, direction_reflected], axis="plot"),
            axis="plot",
            components=("x", "z"),
            label="reflected",
        );
        ax.set_aspect("equal");
        ax.axvline(0, linestyle="dashed", color="black");
        ax.axhspan(ymin=-2, ymax=0, color="lightgray");
        ax.set_ylim(-1, None);
        ax.legend();

    Notes
    -----

    Our goal is to derive a 3D version of Snell's law that can model the
    diffraction from a periodic ruling pattern (diffraction grating).

    To start, consider an incident wave of the form:

    .. math::
        :label: incident-wave

        E_1(\mathbf{r}) = A_1 e^{i \mathbf{k}_1 \cdot \mathbf{r}},

    where :math:`E_1` is the magnitude of the incident electric field,
    :math:`A_1` is the amplitude of the incident wave, and :math:`\mathbf{k}_1` is
    the incident wavevector.

    Now define an interface at :math:`z = 0`, where the index of
    refraction changes, and/or there is a periodic ruling pattern inscribed.
    When the incident wave interacts with this interface, it will create an
    output wave of the form:

    .. math::
        :label: transmitted-wave

        E_2(\mathbf{r}) = A_2 e^{i \mathbf{k}_2 \cdot \mathbf{r}},

    where :math:`E_2` is the magnitude of the output electric field,
    :math:`A_2` is the amplitude of the output wave, and :math:`\mathbf{k}_2` is
    the output wavevector.

    Note in this case we care about only the reflected `or` the transmitted
    wave, not both, since we're only concerned with sequential optics.

    If the interface at :math:`z = 0` `doesn't` have a periodic ruling pattern,
    :math:`E_1` and :math:`E_2` satisfy homogenous Dirichlet boundary conditions.

    .. math::
        :label: boundary-condition

        A_1 \exp\left[i \mathbf{k}_1 \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}) \right]
        = A_2 \exp\left[i \mathbf{k}_2 \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}) \right]

    To include the ruling pattern, we model it as a phase shift of the wave at the interface,

    .. math::
        :label: phase-shift

        \phi(x, y) = i \boldsymbol{\kappa} \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}})

    where the vector

    .. math::

        \boldsymbol{\kappa} = -\frac{2 \pi m}{d} \hat{\boldsymbol{\kappa}},

    :math:`m` is the diffraction order,
    :math:`d` is the groove spacing,
    and :math:`\hat{\boldsymbol{\kappa}}` is a unit vector normal to the
    planes of the rulings.

    With the inclusion of Equation :eq:`phase-shift`, Equation :eq:`boundary-condition` becomes:

    .. math::
        :label: boundary-condition-shifted

         A_1 \exp\left[i (\mathbf{k}_1 + \boldsymbol{\kappa}) \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}) \right]
        = A_2 \exp\left[i \mathbf{k}_2 \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}) \right]

    For Equation :eq:`boundary-condition-shifted` to be true everywhere in
    the :math:`x`-:math:`y` plane, the exponents must be equal:

    .. math::
        :label: boundary-condition-exponents

        (\mathbf{k}_1 + \boldsymbol{\kappa}) \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}})
        = \mathbf{k}_2 \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}).

    If we take :math:`\mathbf{k}_i = k_i \hat{\mathbf{k}}_i = n_i k_0 \hat{\mathbf{k}}_i` for :math:`i=(1,2)`,
    where :math:`k_i` is the incident/output wavenumber,
    :math:`n_i` is the incident/output index of refraction,
    :math:`k_0` is the wavenumber in vacuum,
    and :math:`\hat{\mathbf{k}}_i` is the incident/output propagation direction,
    we get an expression in terms of the output direction, :math:`\hat{\mathbf{k}}_2`:

    .. math::
        :label: boundary-directions

        n_1 (\hat{\mathbf{k}}_1 + \boldsymbol{\kappa} / k_1) \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}})
        = n_2 \hat{\mathbf{k}}_2 \cdot (x \hat{\mathbf{x}} + y \hat{\mathbf{y}}).

    Now, Equation :eq:`boundary-directions` can only be satisfied
    if the components are separately equal since if :math:`y=0`

    .. math::
        :label: k_x

        n_1 (\hat{\mathbf{k}}_1 + \boldsymbol{\kappa} / k_1) \cdot \hat{\mathbf{x}}
        = n_2 \hat{\mathbf{k}}_2 \cdot \hat{\mathbf{x}},

    and if :math:`x=0`

    .. math::
        :label: k_y

        n_1 (\hat{\mathbf{k}}_1 + \boldsymbol{\kappa} / k_1) \cdot \hat{\mathbf{y}}
        = n_2 \hat{\mathbf{k}}_2 \cdot \hat{\mathbf{y}}.

    So, if we define an effective incident propagation direction

    .. math::

        \mathbf{k}_\text{e} = \hat{\mathbf{k}}_1 + \boldsymbol{\kappa} / k_1,

    we can collect Equations :eq:`k_x` and :eq:`k_y` into a single vector
    equation

    .. math::
        :label: snells-law

        n_1 [ \mathbf{k}_\text{e} - ( \mathbf{k}_\text{e} \cdot \hat{\mathbf{n}} ) \hat{\mathbf{n}} ]
        = n_2 [ \hat{\mathbf{k}}_2 - ( \hat{\mathbf{k}}_2 \cdot \hat{\mathbf{n}} ) \hat{\mathbf{n}} ],

    where :math:`\hat{\mathbf{n}} = \hat{\mathbf{z}}` is the vector normal to
    the interface.

    Our goal now is to solve Equation :eq:`snells-law` for
    the output propagation direction, :math:`\hat{\mathbf{k}}_2`,
    and the only unknown is the component of the output propagation direction
    parallel to the surface normal vector, :math:`(\hat{\mathbf{k}}_2 \cdot \hat{\mathbf{n}} )`.
    We can write this in terms of a cross product,

    .. math::
        :label: k2_dot_n

        \hat{\mathbf{k}}_2 \cdot \hat{\mathbf{n}} = \pm \sqrt{1 - |\hat{\mathbf{k}}_2 \times \hat{\mathbf{n}}|^2},

    where the :math:`\pm` represents a transmission or reflection, and the
    cross product :math:`\mathbf{k}_2 \times \hat{\mathbf{n}}` can be found by crossing
    Equation :eq:`snells-law` with :math:`\hat{\mathbf{n}}`,

    .. math::
        :label: k2_cross_n

        \hat{\mathbf{k}}_2 \times \hat{\mathbf{n}}
        = \frac{n_1}{n_2} \mathbf{k}_\text{e} \times \hat{\mathbf{n}},

    which leads to Equation :eq:`k2_dot_n` becoming:

    .. math::
        :label: k2_dot_n_expanded

        \mathbf{k}_2 \cdot \hat{\mathbf{n}}
        &= \pm \sqrt{1 - \left( \frac{n_1}{n_2} \right)^2 |\mathbf{k}_\text{e} \times \hat{\mathbf{n}}|^2} \\
        &= \pm \frac{n_1}{n_2} \sqrt{\left( n_2 / n_1 \right)^2 + (\mathbf{k}_\text{e} \cdot \hat{\mathbf{n}})^2 - k_\text{e}^2 }

    By plugging Equation :eq:`k2_dot_n_expanded` into Equation :eq:`snells-law`,
    and solving for :math:`\hat{\mathbf{k}}_2`, we achieve our goal, an expression for the output ray
    in terms of the input ray and other known quantities.

    .. math::

        \hat{\mathbf{k}}_2
        = \frac{n_1}{n_2} \left[ \mathbf{k}_\text{e}
        + \left(
            \left( -\mathbf{k}_\text{e} \cdot \hat{\mathbf{n}} \right)
            \pm \sqrt{\left( n_2 / n_1 \right)^2 + (\mathbf{k}_\text{e} \cdot \hat{\mathbf{n}})^2 - k_\text{e}^2 }
        \right) \hat{\mathbf{n}} \right]
    """
    a = direction
    n1 = index_refraction
    n2 = index_refraction_new
    m = diffraction_order

    if normal is None:
        normal = na.Cartesian3dVectorArray(0, 0, -1)

    r = n1 / n2

    wavelength_new = wavelength / r

    if np.any(m != 0):
        d = spacing_rulings
        g = normal_rulings
        a = np.where(
            condition=np.isfinite(d),
            x=a - np.sign(-a @ normal) * (m * wavelength_new * g) / (n1 * d),
            y=a,
        )

    c = -a @ normal

    mirror = np.sign(c) * (2 * is_mirror - 1)

    t = np.sqrt(np.square(1 / r) + np.square(c) - np.square(a.length))
    b = r * (a + (c + mirror * t) * normal)

    return b



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika.materials._tests.test_materials import AbstractTestAbstractMaterial


@pytest.mark.parametrize(
    argnames="wavelength,result_expected",
    argvalues=[
        (1.0 * u.eV, 0 * u.electron / u.photon),
        (2.0 * u.eV, 1 * u.electron / u.photon),
        (2 * optika.sensors.energy_electron_hole, 2 * u.electron / u.photon),
    ],
)
def test_quantum_yield_ideal(
    wavelength: u.Quantity | na.AbstractScalar, result_expected: na.AbstractScalar
):
    result = optika.sensors.quantum_yield_ideal(wavelength)
    assert np.all(result == result_expected)


@pytest.mark.parametrize(
    argnames="wavelength",
    argvalues=[
        304 * u.AA,
        na.linspace(100, 200, axis="wavelength", num=4) * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        1,
        0.5,
    ],
)
@pytest.mark.parametrize(
    argnames="n",
    argvalues=[
        1,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_oxide",
    argvalues=[
        10 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_substrate",
    argvalues=[
        10 * u.um,
    ],
)
def test_absorbance(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    n: float | na.AbstractScalar,
    thickness_oxide: u.Quantity | na.AbstractScalar,
    thickness_substrate: u.Quantity | na.AbstractScalar,
):
    result = optika.sensors.absorbance(
        wavelength=wavelength,
        direction=direction,
        n=n,
        thickness_oxide=thickness_oxide,
        thickness_substrate=thickness_substrate,
    )

    assert np.all(result >= 0)
    assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="absorption",
    argvalues=[
        1 / u.mm,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_implant",
    argvalues=[
        1000 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="cce_backsurface",
    argvalues=[
        0.2,
        1,
    ],
)
@pytest.mark.parametrize(
    argnames="cos_incidence",
    argvalues=[
        1,
    ],
)
def test_charge_collection_efficiency(
    absorption: u.Quantity | na.AbstractScalar,
    thickness_implant: u.Quantity | na.AbstractScalar,
    cce_backsurface: u.Quantity | na.AbstractScalar,
    cos_incidence: float | na.AbstractScalar,
):
    result = optika.sensors.charge_collection_efficiency(
        absorption=absorption,
        thickness_implant=thickness_implant,
        cce_backsurface=cce_backsurface,
        cos_incidence=cos_incidence,
    )

    assert np.all(result >= 0)
    assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="wavelength",
    argvalues=[
        304 * u.AA,
        na.linspace(100, 200, axis="wavelength", num=4) * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        None,
        na.Cartesian3dVectorArray(0, 0, 1),
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_oxide",
    argvalues=[
        10 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_implant",
    argvalues=[
        1000 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_substrate",
    argvalues=[
        1 * u.um,
    ],
)
@pytest.mark.parametrize(
    argnames="cce_backsurface",
    argvalues=[
        0.2,
        1,
    ],
)
def test_quantum_efficiency_effective(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    thickness_oxide: u.Quantity | na.AbstractScalar,
    thickness_implant: u.Quantity | na.AbstractScalar,
    thickness_substrate: u.Quantity | na.AbstractScalar,
    cce_backsurface: u.Quantity | na.AbstractScalar,
):
    result = optika.sensors.quantum_efficiency_effective(
        wavelength=wavelength,
        direction=direction,
        thickness_oxide=thickness_oxide,
        thickness_implant=thickness_implant,
        thickness_substrate=thickness_substrate,
        cce_backsurface=cce_backsurface,
    )
    assert np.all(result >= 0)
    assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="iqy",
    argvalues=[1.61 * u.electron / u.photon],
)
@pytest.mark.parametrize(
    argnames="cce",
    argvalues=[0.9],
)
def test_(
    iqy: u.Quantity | na.AbstractScalar,
    cce: float | na.AbstractScalar,
):
    result = optika.sensors.probability_measurement(
        iqy=iqy,
        cce=cce,
    )
    assert np.all(result >= 0)
    assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="photons_expected",
    argvalues=[100 * u.photon],
)
@pytest.mark.parametrize(
    argnames="absorbance",
    argvalues=[0.75],
)
@pytest.mark.parametrize(
    argnames="iqy",
    argvalues=[1.61 * u.electron / u.photon],
)
@pytest.mark.parametrize(
    argnames="absorption",
    argvalues=[1 / u.um],
)
def test_signal(
    photons_expected: u.Quantity | na.AbstractScalar,
    absorbance: float | na.AbstractScalar,
    iqy: u.Quantity | na.AbstractScalar,
    absorption: float | na.AbstractScalar,
):
    result = optika.sensors.signal(
        photons_expected=photons_expected,
        absorbance=absorbance,
        absorption=absorption,
        iqy=iqy,
    )
    assert np.all(result >= 0 * u.electron)


class AbstractTestAbstractImagingSensorMaterial(
    AbstractTestAbstractMaterial,
):
    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                intensity=1e-6 * u.erg,
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_signal(
        self,
        a: optika.sensors.AbstractImagingSensorMaterial,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.signal(rays, normal)
        assert isinstance(result, optika.rays.RayVectorArray)
        assert np.all(result.intensity >= 0 * u.electron)

    @pytest.mark.parametrize(
        argnames="electrons",
        argvalues=[
            100 * u.electron,
        ],
    )
    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            100 * u.AA,
        ],
    )
    @pytest.mark.parametrize(
        argnames="direction",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, 1),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_photons_incident(
        self,
        a: optika.sensors.AbstractImagingSensorMaterial,
        electrons: u.Quantity | na.AbstractScalar,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.photons_incident(
            electrons=electrons,
            wavelength=wavelength,
            direction=direction,
            normal=normal,
        )
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert result.unit.is_equivalent(u.photon)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                intensity=10 * u.electron,
                wavelength=100 * u.AA,
                position=na.Cartesian3dVectorArray() * u.mm,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_charge_diffusion(
        self,
        a: optika.sensors.AbstractImagingSensorMaterial,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.charge_diffusion(rays, normal)
        assert isinstance(result, optika.rays.RayVectorArray)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sensors.IdealImagingSensorMaterial(),
    ],
)
class TestIdealImagingSensorMaterial(
    AbstractTestAbstractImagingSensorMaterial,
):
    pass


class AbstractTestAbstractCCDMaterial(
    AbstractTestAbstractImagingSensorMaterial,
):
    def test_fano_noise(self, a: optika.sensors.AbstractCCDMaterial):
        result = a.fano_noise
        assert np.all(result > 0 * u.electron / u.photon)


class AbstractTestAbstractBackilluminatedCCDMaterial(
    AbstractTestAbstractCCDMaterial,
):
    def test_thickness_oxide(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_oxide
        assert result >= 0 * u.mm

    def test_thickness_implant(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_implant
        assert result >= 0 * u.mm

    def test_thickness_substrate(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_substrate
        assert result >= 0 * u.mm

    def test_cce_backsurface(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.cce_backsurface
        assert result >= 0

    def test_depletion(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.depletion
        assert isinstance(result, optika.sensors.AbstractDepletionModel)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=100 * u.AA,
                position=na.Cartesian3dVectorArray() * u.mm,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_width_charge_diffusion(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.width_charge_diffusion(rays, normal)
        assert np.all(result >= 0 * u.um)

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            500 * u.nm,
        ],
    )
    def test_quantum_yield_ideal(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.quantum_yield_ideal(wavelength)
        assert result >= 0

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_absorbance(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.absorbance(rays, normal)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_charge_collection_efficiency(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.charge_collection_efficiency(rays, normal)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_quantum_efficiency(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.quantum_efficiency(rays, normal)
        assert result > 0 * u.electron / u.photon

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_probability_measurement(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.probability_measurement(rays, normal)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                intensity=100 * u.photon,
                wavelength=100 * u.AA,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_electrons_measured(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.electrons_measured(rays, normal)
        assert isinstance(result, optika.rays.RayVectorArray)
        assert np.all(result.intensity >= 0 * u.electron)


class AbstractTestAbstractStern1994BackilluminatedCCDMaterial(
    AbstractTestAbstractBackilluminatedCCDMaterial,
):
    def test_quantum_efficiency_measured(
        self,
        a: optika.sensors.AbstractStern1994BackilluminatedCCDMaterial,
    ):
        result = a.quantum_efficiency_measured
        assert isinstance(result, na.AbstractFunctionArray)
        assert np.all(result.outputs >= 0)
        assert np.all(result.outputs <= 1.1)
        assert np.all(result.inputs >= 0 * u.nm)



from __future__ import annotations
from typing import TypeVar, Generic
import abc
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    "AbstractRayVectorArray",
    "RayVectorArray",
]

IntensityT = TypeVar("IntensityT", bound=na.ScalarLike)
AttenuationT = TypeVar("AttenuationT", bound=na.ScalarLike)
PositionT = TypeVar("PositionT", bound=na.Cartesian3dVectorArray)
DirectionT = TypeVar("DirectionT", bound=na.Cartesian3dVectorArray)
WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)
IndexRefractionT = TypeVar("IndexRefractionT", bound=na.ScalarLike)
UnvignettedT = TypeVar("UnvignettedT", bound=na.ScalarLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRayVectorArray(
    na.AbstractSpectralPositionalVectorArray,
):
    @property
    @abc.abstractmethod
    def direction(self) -> na.AbstractCartesian3dVectorArray:
        """
        the propagation direction of the ray expressed in
        `direction cosines <https://en.wikipedia.org/wiki/Direction_cosine>`_
        """

    @property
    @abc.abstractmethod
    def intensity(self) -> na.AbstractScalar:
        """
        the radiometric contribution of the ray
        """

    @property
    @abc.abstractmethod
    def attenuation(self) -> na.AbstractScalar:
        """
        the current
        `attenuation coefficient <https://en.wikipedia.org/wiki/Absorption_coefficient>`_
        of the medium that the ray is propagating in
        """

    @property
    @abc.abstractmethod
    def index_refraction(self):
        """
        the current index of refraction of the medium that the ray is
        traveling in
        """

    @property
    @abc.abstractmethod
    def unvignetted(self) -> na.AbstractScalar:
        """
        A boolean mask where :obj:`True` indicates the ray makes it all the
        way through the optical system without being blocked by an aperture,
        and :obj:`False` indicates the ray has been blocked by an aperture.
        """

    @property
    def type_abstract(self) -> type[na.AbstractArray]:
        return AbstractRayVectorArray

    @property
    def type_explicit(self) -> type[na.AbstractExplicitArray]:
        return RayVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError

    @property
    def explicit(self) -> RayVectorArray:
        return super().explicit

    def __array_matmul__(
        self,
        x1: na.ArrayLike,
        x2: na.ArrayLike,
        out: tuple[None | na.AbstractExplicitArray] = (None,),
        **kwargs,
    ) -> na.AbstractExplicitArray | RayVectorArray:
        result = super().__array_matmul__(
            x1=x1,
            x2=x2,
            out=out,
            **kwargs,
        )
        if result is not NotImplemented:
            return result

        if isinstance(x1, AbstractRayVectorArray):
            if isinstance(x2, na.AbstractCartesian3dMatrixArray):
                return dataclasses.replace(
                    x1,
                    position=x1.position @ x2,
                    direction=x1.direction @ x2,
                )
            else:
                return NotImplemented
        elif isinstance(x2, AbstractRayVectorArray):
            if isinstance(x1, na.AbstractCartesian3dMatrixArray):
                return dataclasses.replace(
                    x2,
                    position=x1 @ x2.position,
                    direction=x1 @ x2.direction,
                )
            else:
                return NotImplemented
        else:
            return NotImplemented

    def __array_add__(
        self,
        x1: na.ArrayLike,
        x2: na.ArrayLike,
        out: tuple[None | na.AbstractExplicitArray] = (None,),
        **kwargs,
    ) -> na.AbstractExplicitArray:
        if isinstance(x1, AbstractRayVectorArray):
            if isinstance(x2, na.AbstractCartesian3dVectorArray):
                return dataclasses.replace(
                    x1,
                    position=x1.position + x2,
                )
            else:
                return NotImplemented
        elif isinstance(x2, AbstractRayVectorArray):
            if isinstance(x1, na.AbstractCartesian3dVectorArray):
                return dataclasses.replace(
                    x2,
                    position=x1 + x2.position,
                )
            else:
                return NotImplemented
        else:
            return NotImplemented

    def __array_ufunc__(
        self,
        function: np.ufunc,
        method: str,
        *inputs,
        **kwargs,
    ) -> None | RayVectorArray | tuple[RayVectorArray, ...]:
        result = super().__array_ufunc__(
            function,
            method,
            *inputs,
            **kwargs,
        )
        if result is not NotImplemented:
            return result

        if function is np.add:
            if method == "__call__":
                return self.__array_add__(*inputs, **kwargs)


@dataclasses.dataclass(eq=False, repr=False)
class RayVectorArray(
    AbstractRayVectorArray,
    na.SpectralPositionalVectorArray[WavelengthT, PositionT],
    Generic[
        IntensityT,
        PositionT,
        DirectionT,
        WavelengthT,
        IndexRefractionT,
        UnvignettedT,
    ],
):
    direction: DirectionT = 0
    intensity: IntensityT = 1
    attenuation: AttenuationT = 0 / u.mm
    index_refraction: IndexRefractionT = 1

    unvignetted: na.AbstractScalar = True
    """
    A boolean mask where :obj:`True` indicates the ray makes it all the
    way through the optical system without being blocked by an aperture,
    and :obj:`False` indicates the ray has been blocked by an aperture.
    """

    @classmethod
    def from_scalar(
        cls,
        scalar: na.AbstractScalar,
        like: None | na.AbstractExplicitVectorArray = None,
    ) -> RayVectorArray:
        return cls(
            wavelength=scalar,
            position=scalar,
            direction=scalar,
            intensity=scalar,
            attenuation=scalar,
            index_refraction=scalar,
            unvignetted=scalar,
        )



from __future__ import annotations
import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika
from ._layers import Layer

__all__ = [
    "AbstractMaterial",
    "Vacuum",
    "AbstractMirror",
    "Mirror",
    "MeasuredMirror",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMaterial(
    optika.mixins.Printable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
):
    @abc.abstractmethod
    def index_refraction(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        """
        the index of refraction of this material for the given input rays

        Parameters
        ----------
        rays
            input rays used to evaluate the index of refraction
        """

    @abc.abstractmethod
    def attenuation(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        """
        the attenuation coefficient of the given rays

        Parameters
        ----------
        rays
            input rays to calculate the attenuation coefficient for
        """

    @abc.abstractmethod
    def efficiency(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        """
        The fraction of light that passes through the interface.

        Parameters
        ----------
        rays
            the input rays to calculate the efficiency for
        normal
            the vector perpendicular to the optical surface
        """

    @property
    @abc.abstractmethod
    def is_mirror(self) -> bool:
        """
        flag controlling whether this material reflects or transmits light
        """


@dataclasses.dataclass(eq=False, repr=False)
class Vacuum(
    AbstractMaterial,
):
    @property
    def shape(self) -> dict[str, int]:
        return dict()

    @property
    def transformation(self) -> None:
        return None

    def index_refraction(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 1

    def attenuation(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 0 / u.mm

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        return 1

    @property
    def is_mirror(self) -> bool:
        return False


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMirror(
    AbstractMaterial,
):

    @property
    @abc.abstractmethod
    def substrate(self) -> None | Layer:
        """
        A layer representing the substrate supporting the reflective surface.
        """

    @property
    def transformation(self) -> None:
        return None

    def index_refraction(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return rays.index_refraction

    def attenuation(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return rays.attenuation

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        return 1

    @property
    def is_mirror(self) -> bool:
        return True


@dataclasses.dataclass(eq=False, repr=False)
class Mirror(
    AbstractMirror,
):
    substrate: None | Layer = None
    """A layer representing the substrate supporting the reflective surface."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.substrate),
        )


@dataclasses.dataclass(eq=False, repr=False)
class MeasuredMirror(
    AbstractMirror,
):
    """
    A mirror where the reflectivity has been measured by an external source as
    a function of wavelength.

    Examples
    --------

    Create a mirror where the reflectivity is a Gaussian centered
    at 304 Angstroms.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define the mean and standard deviation of the reflectivity peak.
        center = 304 * u.AA
        width = 10 * u.AA

        # Define a grid of wavelengths
        wavelength_min = center - 3 * width
        wavelength_max = center + 3 * width
        wavelength = na.linspace(
            start=wavelength_min,
            stop=wavelength_max,
            axis="wavelength",
            num=11,
        )

        # Define an array of simulated reflectivity measurements
        efficiency = na.FunctionArray(
            inputs=na.SpectralDirectionalVectorArray(
                wavelength=wavelength,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            outputs=np.exp(-np.square((wavelength - center) / width) / 2),
        )

        # Create an instance of a MeasuredMirror object
        mirror = optika.materials.MeasuredMirror(efficiency)

        # Define a new grid of wavelengths at which to evaluate the interpolated
        # reflectivity
        wavelength_interp = na.linspace(
            start=wavelength_min,
            stop=wavelength_max,
            axis="wavelength",
            num=1001,
        )

        # Evaluate the interpolated reflectivity
        efficiency_interp = mirror.efficiency(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_interp,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, 1),
        )

        # Plot the interpolated reflectivity vs the measured reflectivity
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.plot(wavelength_interp, efficiency_interp, label="interpolated");
            na.plt.scatter(efficiency.inputs.wavelength, efficiency.outputs, label="measured");
            ax.set_xlabel(f"wavelength ({ax.xaxis.get_label().get_text()})");
            ax.set_ylabel(f"reflectivity");
            ax.legend();
    """

    efficiency_measured: na.FunctionArray[
        na.SpectralDirectionalVectorArray, na.AbstractScalar
    ] = dataclasses.MISSING
    """
    A function array that maps wavelengths and incidence angles to the
    measured reflectivity.
    """

    substrate: None | Layer = None
    """A layer representing the substrate supporting the reflective surface."""

    serial_number: None | str | na.AbstractArray = None
    """A unique number associated with this material"""

    @property
    def shape(self) -> dict[str, int]:
        axis_wavelength = self.efficiency_measured.inputs.wavelength.axes
        shape = optika.shape(self.efficiency_measured.outputs)
        for ax in axis_wavelength:
            shape.pop(ax, None)
        return na.broadcast_shapes(
            shape,
            optika.shape(self.substrate),
            optika.shape(self.serial_number),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:

        measurement = self.efficiency_measured

        wavelength = measurement.inputs.wavelength
        direction = measurement.inputs.direction
        efficiency = measurement.outputs

        if direction.size != 1:  # pragma: nocover
            raise ValueError(
                "Interpolating over different incidence angles is not supported."
            )

        if wavelength.ndim != 1:  # pragma: nocover
            raise ValueError(
                f"wavelength must be one dimensional, got shape {wavelength.shape}"
            )

        return na.interp(
            x=rays.wavelength,
            xp=wavelength,
            fp=efficiency,
        )



from typing import Any
import numpy as np
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "shape",
    "direction",
    "angles",
]


_shaped_type = (
    na.AbstractArray,
    na.transformations.AbstractTransformation,
    optika.mixins.Shaped,
)


def shape(a: Any) -> dict[str, int]:
    """
    Return the array shape of the given object.

    If the given object is an instance of :class:`named_arrays.AbstractArray`,
    :class:`named_arrays.transformations.AbstractTransformation`, or
    :class:`optika.mixins.Shaped`, ``a.shape`` will be returned.
    Otherwise an empty dictionary will be returned.

    Parameters
    ----------
    a
        The object to find the shape of.
    """
    if isinstance(a, _shaped_type):
        return a.shape
    else:
        return dict()


def direction(
    angles: na.AbstractCartesian2dVectorArray,
) -> na.Cartesian3dVectorArray:
    r"""
    Given a 2D vector of azimuth and elevation angles, convert to a
    3D vector of direction cosines.

    Parameters
    ----------
    angles
        A vector of azimuth and elevation angles.

    Notes
    -----
    If the azimuth and elevation angles are taken to :math:`\phi_x` and
    :math:`\phi_y`, the direction cosines :math:`\vec{d}` can be found by

    .. math::

        \vec{d} = R_y(\phi_x) R_x(\phi_y) \hat{z}

    where :math:`R_x(\theta)` and :math:`R_y(\theta)` are the rotation matrices
    about the :math:`x` and :math:`y` axes respectively.

    See Also
    --------
    :func:`angles` : Inverse of this function
    """
    return na.Cartesian3dVectorArray(
        x=-np.cos(angles.y) * np.sin(angles.x),
        y=-np.sin(angles.y),
        z=+np.cos(angles.y) * np.cos(angles.x),
    )


def angles(
    direction: na.AbstractCartesian3dVectorArray,
) -> na.Cartesian2dVectorArray:
    """
    Convert a 3D vector of direction cosines to a 2D vector of azimuth and
    elevation angles.

    Parameters
    ----------
    direction
        A vector of direction cosines.

    See Also
    --------
    :func:`direction` : Inverse of this function
    """
    if na.unit(direction) is None:
        direction = direction << u.dimensionless_unscaled
    return na.Cartesian2dVectorArray(
        x=-np.arctan2(direction.x, direction.z).to(u.deg),
        y=-np.arcsin(direction.y / direction.length).to(u.deg),
    )



import pytest
import optika
from .._materials_test import AbstractTestAbstractStern1994BackilluminatedCCDMaterial


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sensors.TektronixTK512CBMaterial(),
    ],
)
class TestTektronixTK512CBMaterial(
    AbstractTestAbstractStern1994BackilluminatedCCDMaterial,
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests import test_mixins


class AbstractTestAbstractRoughnessParameters(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_period_min(self, a: optika.metrology.AbstractRoughnessParameters):
        assert np.issubdtype(na.get_dtype(a.period_min), float)
        assert na.unit_normalized(a.period_min).is_equivalent(u.mm)

    def test_period_max(self, a: optika.metrology.AbstractRoughnessParameters):
        assert np.issubdtype(na.get_dtype(a.period_max), float)
        assert na.unit_normalized(a.period_max).is_equivalent(u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.metrology.RoughnessParameters(
            period_min=2 * u.mm,
            period_max=4 * u.mm,
        )
    ],
)
class TestRoughnessParameters(
    AbstractTestAbstractRoughnessParameters,
):
    pass



import pytest
import abc
import numpy as np
import optika.propagators
import optika.rays._tests.test_ray_vectors


class AbstractTestAbstractPropagator(
    abc.ABC,
):
    pass


class AbstractTestAbstractRayPropagator(
    AbstractTestAbstractPropagator,
):
    @pytest.mark.parametrize("rays", optika.rays._tests.test_ray_vectors.rays)
    def test_propagate_rays(
        self,
        a: optika.propagators.AbstractRayPropagator,
        rays: optika.rays.AbstractRayVectorArray,
    ):
        result = a.propagate_rays(rays)

        assert isinstance(result, optika.rays.AbstractRayVectorArray)
        assert not np.all(result == rays)


class AbstractTestAbstractLightPropagator(
    AbstractTestAbstractRayPropagator,
):
    pass



import pathlib
import numpy as np
import astropy.units as u
import named_arrays as na
from .._depletion import E2VCCD64ThickDepletionModel
from .._materials import AbstractStern1994BackilluminatedCCDMaterial

__all__ = [
    "E2VCCD97Material",
]


class E2VCCD97Material(
    AbstractStern1994BackilluminatedCCDMaterial,
):
    """
    A model of the light-sensitive material of an e2v CCD97 sensor based on
    measurements by :cite:t:`Moody2017` and :cite:t:`Heymes2020`.

    This is a measurement of e2v's "enhanced" process, which has a narrower
    partial charge collection region than e2v's "standard" process.

    This model uses the :class:`optika.sensors.E2VCCD64ThickDepletionModel`
    to represent the depletion region.

    Examples
    --------

    Plot the measured E2VCCD97 quantum efficiency vs the fitted
    quantum efficiency calculated using the method of :cite:t:`Stern1994`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Create a new instance of the e2v CCD97 light-sensitive material
        material = optika.sensors.E2VCCD97Material()

        # Store the wavelengths at which the QE was measured
        wavelength_measured = material.quantum_efficiency_measured.inputs

        # Store the QE measurements
        qe_measured = material.quantum_efficiency_measured.outputs

        # Define a grid of wavelengths with which to evaluate the fitted QE
        wavelength_fit = na.geomspace(5, 10000, axis="wavelength", num=1001) * u.AA

        # Evaluate the fitted QE using the given wavelengths
        qe_fit = material.quantum_efficiency_effective(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the measured QE vs the fitted QE
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.scatter(
                wavelength_measured,
                qe_measured,
                label="measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit,
                label="fit",
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({wavelength_fit.unit:latex_inline})")
            ax.set_ylabel("quantum efficiency")
            ax.legend()

    The thickness of the oxide layer found by the fit is

    .. jupyter-execute::

        material.thickness_oxide

    The thickness of the implant layer found by the fit is

    .. jupyter-execute::

        material.thickness_implant

    The thickness of the substrate is

    .. jupyter-execute::

        material.thickness_substrate

    The differential charge collection efficiency at the backsurface
    found by the fit is

    .. jupyter-execute::

        material.cce_backsurface

    And the roughness of the substrate found by the fit is

    .. jupyter-execute::

        material.roughness_substrate

    |

    Now plot the effective quantum efficiency of the fit to this data vs. the fit
    to the data in :class:`optika.sensors.E2VCCD203Material`

    .. jupyter-execute::

        material_ccd_aia = optika.sensors.E2VCCD203Material()

        qe_fit_aia = material_ccd_aia.quantum_efficiency_effective(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.scatter(
                wavelength_measured,
                qe_measured,
                label="Heymes 2020 measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit,
                label="Heymes 2020 fit",
            )
            na.plt.scatter(
                material_ccd_aia.quantum_efficiency_measured.inputs,
                material_ccd_aia.quantum_efficiency_measured.outputs,
                label="Boerner 2012 measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit_aia,
                label="Boerner 2012 fit",
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({wavelength_fit.unit:latex_inline})")
            ax.set_ylabel("quantum efficiency")
            ax.legend()

    |

    Plot the width of the charge diffusion kernel for this sensor as a function
    of wavelength.

    .. jupyter-execute::

        # Compute the width of the charge diffusion kernel
        # for each wavelength.
        width = material.width_charge_diffusion(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the results
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.plot(
                wavelength_fit,
                width,
                ax=ax,
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax.set_ylabel(f"width ({ax.get_ylabel()})")
    """

    @property
    def quantum_efficiency_measured(self) -> na.FunctionArray:

        directory = pathlib.Path(__file__).parent

        energy_moody, qe_moody = np.genfromtxt(
            fname=directory / "e2v_ccd97_qe_moody2017.csv",
            delimiter=", ",
            unpack=True,
        )
        energy_moody = energy_moody << u.eV
        wavelength_moody = energy_moody.to(u.AA, equivalencies=u.spectral())
        qe_moody = qe_moody << u.percent

        wavelength_heymes, qe_heymes = np.genfromtxt(
            fname=directory / "e2v_ccd97_qe_heymes2020.csv",
            delimiter=", ",
            unpack=True,
        )
        wavelength_heymes = wavelength_heymes << u.nm
        qe_heymes = qe_heymes << u.percent

        wavelength = np.concatenate([wavelength_moody, wavelength_heymes])
        qe = np.concatenate([qe_moody, qe_heymes]).to_value(u.dimensionless_unscaled)

        return na.FunctionArray(
            inputs=na.ScalarArray(wavelength, axes="wavelength"),
            outputs=na.ScalarArray(qe, axes="wavelength"),
        )

    @property
    def thickness_substrate(self) -> u.Quantity:
        return 14 * u.um

    @property
    def depletion(self) -> E2VCCD64ThickDepletionModel:
        return E2VCCD64ThickDepletionModel()

    @property
    def shape(self) -> dict[str, int]:
        return dict()



"""
A collection of transfer matrices for computing the efficiency of optical films and
multilayers.
"""

from __future__ import annotations
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from . import snells_law


__all__ = [
    "refraction",
    "propagation",
    "transfer",
]


def refraction(
    wavelength: u.Quantity | na.AbstractScalar,
    direction_left: float | na.AbstractScalar,
    direction_right: float | na.AbstractScalar,
    polarized_s: bool | na.AbstractScalar,
    n_left: float | na.AbstractScalar,
    n_right: float | na.AbstractScalar,
    interface: None | optika.materials.profiles.AbstractInterfaceProfile = None,
) -> na.Cartesian2dMatrixArray:
    r"""
    Compute the refractive matrix, which represents the bending of light
    at an interface.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum
    direction_left
        The component of the incident light's propagation direction before
        the interface antiparallel to the surface normal.
    direction_right
        The component of the incident light's propagation direction after
        the interface antiparallel to the surface normal.
    polarized_s
        If :obj:`True`, the incident light is :math:`s`-polarized.
        If :obj:`False`, the incident light is :math:`p`-polarized.
    n_left
        The complex index of refraction on the left side of the interface.
    n_right
        The complex index of refraction on the right side of the interface.
    interface
        The interface profile between the left side and the right side.

    Examples
    --------

    Compute the refractive matrix for :math:`s`-polarized light normally
    incident on the interface between vacuum and silicon dioxide.

    .. jupyter-execute::

        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the wavelength of the incident light
        wavelength = 100 * u.AA

        # Initialize a representation of silicon dioxide
        sio2 = optika.chemicals.Chemical("SiO2")

        # Compute the refractive matrix
        optika.materials.matrices.refraction(
            wavelength=wavelength,
            direction_left=1,
            direction_right=1,
            polarized_s=True,
            n_left=1,
            n_right=sio2.n(wavelength),
        )

    Notes
    -----

    The refractive matrix is given by :cite:t:`Yeh1988` Equation 5.1-12,

    .. math::
        :label: refractive-matrix

        W_{kij} = \frac{1}{t_{kij}} \begin{pmatrix}
                                    1 & r_{kij} \\
                                    r_{kij} & 1 \\
                                  \end{pmatrix},

    where :math:`k=(s, p)` is the polarization state, :math:`i=j-1` is the index
    of the previous material, :math:`j` is the index of the current material,

    .. math::
        :label: fresnel-reflection

        r_{kij} = \frac{q_{ki} - q_{kj}}{q_{ki} + q_{kj}}

    is the Fresnel reflection coefficient between materials :math:`i` and :math:`j`,

    .. math::
        :label: fresnel-transmission

        t_{kij} = \frac{2 q_{ki}}{q_{ki} + q_{kj}}

    is the Fresnel transmission coefficient between materials :math:`i` and :math:`j`,

    .. math::

        q_{si} = n_i \cos \theta_i

    and

    .. math::

        q_{pi} = \frac{\cos \theta_i}{n_i}

    are the :math:`z` components of the wave's momentum for :math:`s` and
    :math:`p` polarization,
    :math:`n_i` is the index of refraction inside material :math:`i`,
    and :math:`\theta_i` is the angle between the wave's propagation direction
    and the vector normal to the interface inside material :math:`i`.
    """
    direction_i = direction_left
    direction_j = direction_right

    direction_i = np.where(polarized_s, direction_i, np.conj(direction_i))
    direction_j = np.where(polarized_s, direction_j, np.conj(direction_j))

    n_i = n_left
    n_j = n_right

    impedance_i = np.where(polarized_s, n_i, 1 / n_i)
    impedance_j = np.where(polarized_s, n_j, 1 / n_j)

    q_i = direction_i * impedance_i
    q_j = direction_j * impedance_j

    a_ij = q_i + q_j

    r_ij = (q_i - q_j) / a_ij
    t_ij = 2 * q_i / a_ij

    if interface is not None:
        r_ij = r_ij * interface.reflectivity(
            wavelength=wavelength,
            direction=direction_i,
            n=n_i,
        )

    result = na.Cartesian2dMatrixArray(
        x=na.Cartesian2dVectorArray(1, r_ij),
        y=na.Cartesian2dVectorArray(r_ij, 1),
    )

    result = result / t_ij

    return result


def propagation(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    thickness: u.Quantity | na.AbstractScalar,
    n: u.Quantity | na.AbstractScalar,
) -> na.Cartesian2dMatrixArray:
    r"""
    Compute the propagation matrix, which propagates the electric field
    through a homogenous slab.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The component of the incident light's propagation direction
        antiparallel to the surface normal.
    thickness
        The thickness of the material.
    n
        The complex index of refraction of the material

    Examples
    --------

    Compute the propagation matrix for :math:`s`-polarized light normally
    incident on a layer of silicon dioxide

    .. jupyter-execute::

        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the wavelength of the incident light
        wavelength = 100 * u.AA

        # Initialize a representation of silicon dioxide
        sio2 = optika.chemicals.Chemical("SiO2")

        # Compute the propagation matrix
        optika.materials.matrices.propagation(
            wavelength=wavelength,
            direction=1,
            thickness=10 * u.nm,
            n=sio2.n(wavelength),
        )

    Notes
    -----

    The propagation matrix for a homogenous slab is given by
    :cite:t:`Yeh1988` Equation 5.1-24,

    .. math::
        :label: propagation-matrix

        U = \begin{pmatrix}
                e^{-i \beta} & 0 \\
                0 & e^{i \beta} \\
            \end{pmatrix},

    where

    .. math::
        :label: propagation-phase

        \beta = \frac{2 \pi}{\lambda} n h \cos \theta

    is the phase change from propagating through the material,
    :math:`n` is the index of refraction inside the material,
    :math:`\lambda` is the wavelength of the incident light in vacuum,
    :math:`h` is the thickness of the material,
    and :math:`\theta` is the angle between the surface normal and the propagation
    direction of the incident light.
    """

    beta = 2 * np.pi * thickness * n * direction / wavelength

    return na.Cartesian2dMatrixArray(
        x=na.Cartesian2dVectorArray(np.exp(-1j * beta), 0),
        y=na.Cartesian2dVectorArray(0, np.exp(+1j * beta)),
    )


def transfer(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    polarized_s: bool | na.ScalarArray,
    thickness: u.Quantity | na.AbstractScalar,
    n: float | na.AbstractScalar,
    normal: na.AbstractCartesian3dVectorArray,
) -> na.Cartesian2dMatrixArray:
    r"""
    Compute the transfer matrix for a homogenous slab of material using
    :func:`refraction` and :func:`propagation`.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The propagation direction of the incident light in vacuum.
    polarized_s
        If :obj:`True`, the incident light is :math:`s`-polarized.
        If :obj:`False`, the incident light is :math:`p`-polarized.
    thickness
        The thickness of the homogenous slab.
    n
        The index of refraction of the material
    normal
        The vector perpendicular to the surface of the slab.

    Examples
    --------

    Compute the transfer matrix of a 10-nm-thick slab of silicon dioxide for
    normally-incident :math:`s`-polarized light.

    .. jupyter-execute::

        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the wavelength of the incident light
        wavelength = 100 * u.AA

        # Initialize a representation of silicon dioxide
        sio2 = optika.chemicals.Chemical("SiO2")

        # Compute the transfer matrix
        optika.materials.matrices.transfer(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(0, 0, 1),
            polarized_s=True,
            thickness=10 * u.nm,
            n=sio2.n(wavelength),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

    Notes
    -----

    If :math:`W_{kij}` is the refractive matrix for the interface on the left side
    of the slab (computed using :func:`refraction`),
    :math:`U_{kj}` is the propagation matrix for the slab (computed using
    :func:`propagation`), and :math:`W_{kji}` is the refractive matrix
    for the right side of the slab, then the transfer matrix for the slab can
    be computed using the product of these three matrices:

    .. math::
        :label: transfer-matrix

        T_{kj} = W_{kij} U_{kj} W_{kji}
    """

    direction_internal = snells_law(
        wavelength=wavelength,
        direction=direction,
        index_refraction=1,
        index_refraction_new=np.real(n),
        normal=normal,
    )

    direction = -direction @ normal
    direction_internal = -direction_internal @ normal

    matrix_refractive_left = refraction(
        wavelength=wavelength,
        direction_left=direction,
        direction_right=direction_internal,
        polarized_s=polarized_s,
        n_left=1,
        n_right=n,
    )

    matrix_propagation = propagation(
        wavelength=wavelength,
        direction=direction_internal,
        thickness=thickness,
        n=n,
    )

    matrix_refractive_right = refraction(
        wavelength=wavelength,
        direction_left=direction_internal,
        direction_right=direction,
        polarized_s=polarized_s,
        n_left=n,
        n_right=1,
    )

    return matrix_refractive_left @ matrix_propagation @ matrix_refractive_right



import pytest
import numpy as np
import matplotlib
import astropy.units as u
import astropy.visualization
import named_arrays as na
import optika._tests.test_mixins


class AbstractTestAbstractLayer(
    optika._tests.test_mixins.AbstractTestPrintable,
    optika._tests.test_mixins.AbstractTestShaped,
):

    @pytest.mark.parametrize("wavelength", [100 * u.AA])
    def test_n(
        self,
        a: optika.materials.AbstractLayer,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.n(wavelength)
        assert np.all(np.real(result) >= 0)
        assert np.all(np.imag(result) >= 0)

    def test_thickness(
        self,
        a: optika.materials.AbstractLayer,
    ):
        result = a.thickness
        assert np.all(result >= 0 * u.nm)

    def test_interface(
        self,
        a: optika.materials.AbstractLayer,
    ):
        result = a.interface
        if result is not None:
            assert isinstance(
                result, optika.materials.profiles.AbstractInterfaceProfile
            )

    def test_layer_sequence(
        self,
        a: optika.materials.AbstractLayer,
    ):
        result = a.layer_sequence
        assert isinstance(result, optika.materials.LayerSequence)
        for layer in result.layers:
            assert isinstance(layer, optika.materials.AbstractLayer)

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            na.geomspace(100, 1000, axis="wavelength", num=4) * u.AA,
        ],
    )
    @pytest.mark.parametrize("direction", [1])
    @pytest.mark.parametrize("polarized_s", [False, True])
    @pytest.mark.parametrize("n", [1])
    class TestMatrixTransfer:
        def test_transfer(
            self,
            a: optika.materials.AbstractLayer,
            wavelength: u.Quantity | na.AbstractScalar,
            direction: float | na.AbstractScalar,
            polarized_s: bool | na.AbstractScalar,
            n: float | na.AbstractScalar,
        ):
            n, direction, result, where = a.transfer(
                wavelength=wavelength,
                direction=direction,
                polarized_s=polarized_s,
                n=n,
            )
            assert np.all(n > 0)
            assert isinstance(direction, na.AbstractScalar)
            assert isinstance(result, na.AbstractCartesian2dMatrixArray)
            assert np.all(result.determinant != 0)
            assert np.all(np.isfinite(result))

    @pytest.mark.parametrize("z", [0 * u.nm])
    @pytest.mark.parametrize("ax", [None])
    def test_plot(
        self,
        a: optika.materials.AbstractLayer,
        z: u.Quantity,
        ax: None | matplotlib.axes.Axes,
    ):
        with astropy.visualization.quantity_support():
            result = a.plot(
                z=z,
                ax=ax,
            )

        assert isinstance(result, list)
        for item in result:
            assert isinstance(item, matplotlib.patches.Rectangle)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.Layer(
            chemical="Si",
            thickness=10 * u.nm,
            kwargs_plot=dict(
                color="tab:orange",
                alpha=0.5,
            ),
        ),
        optika.materials.Layer(
            chemical="Si",
            thickness=10 * u.nm,
            x_label=-0.1,
        ),
        optika.materials.Layer(
            chemical="Si",
            thickness=10 * u.nm,
            x_label=1.1,
        ),
    ],
)
class TestLayer(
    AbstractTestAbstractLayer,
):
    pass


class AbstractTestAbstractLayerSequence(
    AbstractTestAbstractLayer,
):
    def test_layers(self, a: optika.materials.AbstractLayerSequence):
        result = a.layers
        for layer in result:
            assert isinstance(layer, optika.materials.AbstractLayer)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.LayerSequence(
            layers=[
                optika.materials.Layer(
                    chemical="SiO2",
                    thickness=10 * u.nm,
                ),
                optika.materials.Layer(
                    chemical="Si",
                    thickness=1 * u.um,
                ),
            ]
        ),
        optika.materials.LayerSequence(
            layers=[
                optika.materials.Layer(
                    chemical="SiO2",
                    thickness=10 * u.nm,
                ),
                optika.materials.LayerSequence(
                    [
                        optika.materials.Layer(
                            chemical="Si",
                            thickness=1 * u.um,
                        ),
                        optika.materials.Layer(
                            chemical="Mo",
                            thickness=1 * u.um,
                        ),
                    ]
                ),
            ]
        ),
    ],
)
class TestLayerSequence(
    AbstractTestAbstractLayerSequence,
):

    @pytest.mark.parametrize("item", [0, slice(0, 1)])
    def test__getitem__(
        self,
        a: optika.materials.LayerSequence,
        item: int | slice,
    ):
        result = a[item]
        if isinstance(item, int):
            assert isinstance(result, optika.materials.AbstractLayer)
        else:
            assert isinstance(result, optika.materials.LayerSequence)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.PeriodicLayerSequence(
            layers=[
                optika.materials.Layer(
                    chemical="Si",
                    thickness=10 * u.nm,
                    interface=optika.materials.profiles.ErfInterfaceProfile(
                        width=2 * u.nm,
                    ),
                ),
                optika.materials.Layer(
                    chemical="Mo",
                    thickness=10 * u.nm,
                    interface=optika.materials.profiles.ErfInterfaceProfile(
                        width=2 * u.nm,
                    ),
                ),
            ],
            num_periods=10,
        ),
        optika.materials.PeriodicLayerSequence(
            layers=[
                optika.materials.PeriodicLayerSequence(
                    layers=[
                        optika.materials.Layer(
                            chemical="Si",
                            thickness=10 * u.nm,
                            interface=optika.materials.profiles.ErfInterfaceProfile(
                                width=2 * u.nm,
                            ),
                        ),
                        optika.materials.Layer(
                            chemical="Mo",
                            thickness=10 * u.nm,
                            interface=optika.materials.profiles.ErfInterfaceProfile(
                                width=2 * u.nm,
                            ),
                        ),
                    ],
                    num_periods=2,
                )
            ],
            num_periods=5,
        ),
    ],
)
class TestPeriodicLayerSequence(
    AbstractTestAbstractLayerSequence,
):
    class TestMatrixTransfer(
        AbstractTestAbstractLayerSequence.TestMatrixTransfer,
    ):
        def test_transfer(
            self,
            a: optika.materials.PeriodicLayerSequence,
            wavelength: u.Quantity | na.AbstractScalar,
            direction: float | na.AbstractScalar,
            polarized_s: bool | na.AbstractScalar,
            n: float | na.AbstractScalar,
        ):
            super().test_transfer(
                a=a,
                wavelength=wavelength,
                direction=direction,
                polarized_s=polarized_s,
                n=n,
            )

            b = optika.materials.LayerSequence(list(a.layers) * a.num_periods)

            (
                n_test,
                direction_test,
                result_test,
                where_test,
            ) = a.transfer(
                wavelength=wavelength,
                direction=direction,
                polarized_s=polarized_s,
                n=n,
            )

            (
                n_expected,
                direction_expected,
                result_expected,
                where_expected,
            ) = b.transfer(
                wavelength=wavelength,
                direction=direction,
                polarized_s=polarized_s,
                n=n,
            )

            assert np.allclose(n_test, n_expected)
            assert np.allclose(direction_test, direction_expected)
            assert np.allclose(result_test, result_expected)
            assert np.allclose(where_test, where_expected)



"""
Models of the depletion region of a semiconducting imaging sensor.
"""

from ._depletion import (
    AbstractDepletionModel,
    AbstractJanesickDepletionModel,
)
from ._e2v_ccd64_thick import E2VCCD64ThickDepletionModel
from ._e2v_ccd64_thin import E2VCCD64ThinDepletionModel

__all__ = [
    "AbstractDepletionModel",
    "AbstractJanesickDepletionModel",
    "E2VCCD64ThickDepletionModel",
    "E2VCCD64ThinDepletionModel",
]



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests import test_mixins


class AbstractTestAbstractDepletionModel(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_thickness(
        self,
        a: optika.sensors.AbstractDepletionModel,
    ):
        result = a.thickness
        assert np.all(result > 0 * u.um)


class AbstractTestAbstractJanesickDepletionModel(
    AbstractTestAbstractDepletionModel,
):

    def test_chemical_substrate(
        self,
        a: optika.sensors.AbstractJanesickDepletionModel,
    ):
        result = a.chemical_substrate
        assert isinstance(result, optika.chemicals.AbstractChemical)

    def test_thickness_substrate(
        self,
        a: optika.sensors.AbstractJanesickDepletionModel,
    ):
        result = a.thickness_substrate
        assert np.all(result > 0 * u.um)

    def test_width_pixel(
        self,
        a: optika.sensors.AbstractJanesickDepletionModel,
    ):
        result = a.width_pixel
        assert np.all(result > 0 * u.um)

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            100 * u.AA,
            na.geomspace(1, 10000, "w", num=11) * u.AA,
        ],
    )
    def test_mean_charge_capture(
        self,
        a: optika.sensors.AbstractJanesickDepletionModel,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.mean_charge_capture(wavelength)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    def test_mean_charge_capture_measured(
        self,
        a: optika.sensors.AbstractJanesickDepletionModel,
    ):
        result = a.mean_charge_capture_measured
        assert isinstance(result, na.AbstractFunctionArray)
        assert np.all(result.inputs > 0 * u.AA)
        assert np.all(result.outputs > 0)



from __future__ import annotations
from typing import Sequence
import abc
import dataclasses
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika
from . import matrices, snells_law_scalar

__all__ = [
    "AbstractLayer",
    "Layer",
    "AbstractLayerSequence",
    "LayerSequence",
    "PeriodicLayerSequence",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLayer(
    optika.mixins.Printable,
    optika.mixins.Shaped,
    abc.ABC,
):
    """
    An interface for representing a single homogeneous layer or a sequence
    of homogeneous layers.
    """

    @abc.abstractmethod
    def n(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The index of refraction on the left side of the layer.

        Parameters
        ----------
        wavelength
            The wavelength of the incident light in vacuum
        """

    @property
    @abc.abstractmethod
    def thickness(self) -> u.Quantity | na.AbstractScalar:
        """The thickness of this layer."""

    @property
    @abc.abstractmethod
    def _thickness_plot(self) -> u.Quantity | na.AbstractScalar:
        """The thickness of the layer when plotted."""

    @property
    @abc.abstractmethod
    def interface(self) -> None | optika.materials.profiles.AbstractInterfaceProfile:
        """
        The interface profile on the left side of this layer.
        """

    @property
    @abc.abstractmethod
    def layer_sequence(self) -> LayerSequence:
        """
        This object interpreted as an instance of :class:`LayerSequence`.
        """

    @abc.abstractmethod
    def transfer(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        polarized_s: bool | na.AbstractScalar,
        n: float | na.AbstractScalar,
        where: bool | na.AbstractScalar = True,
    ) -> tuple[
        na.AbstractScalar,
        na.AbstractScalar,
        na.Cartesian2dMatrixArray,
        na.AbstractScalar,
    ]:
        """
        Compute the index of refraction, internal propagation direction,
        transfer matrix, and where the transfer matrix is valid for this layer,
        which propagates the electric field from the left side of the layer to
        the right side.

        Parameters
        ----------
        wavelength
            The wavelength of the incident light in vacuum.
        direction
            The component of the incident light's propagation direction in the
            medium before this layer antiparallel to the surface normal.
        polarized_s
            If :obj:`True`, the incident light is :math:`s`-polarized.
            If :obj:`False`, the incident light is :math:`p`-polarized.
        n
            The complex index of refraction of the medium before this layer.
        where
            The boolean mask that should be applied to the transfer matrix.
            :obj:`True` where the transfer matrix is invertible.
        """

    def plot(
        self,
        x: u.Quantity = 0 * u.nm,
        z: u.Quantity = 0 * u.nm,
        width: u.Quantity = 10 * u.nm,
        ax: None | matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:
        """
        Plot this layer.

        Parameters
        ----------
        x
            The horizontal offset of the plotted layer
        z
            The vertical offset of the plotted layer.
        width
            The horizontal width of the plotted layer.
        ax
            The matplotlib axes instance on which to plot the layer.
        kwargs
            Additional keyword arguments to pass into.
        """


@dataclasses.dataclass(eq=False, repr=False)
class Layer(
    AbstractLayer,
):
    """
    An isotropic, homogenous layer of optical material.

    Examples
    --------

    Plot a 10-nm-thick layer of silicon.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import optika

        # Create a layer of silicon
        layer = optika.materials.Layer(
            chemical="Si",
            thickness=10 * u.nm,
            kwargs_plot=dict(
                color="tab:blue",
                alpha=0.5,
            ),
        )

        # Plot the layer
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            layer.plot(ax=ax)
            ax.set_axis_off()
            ax.autoscale_view()
    """

    chemical: None | str | optika.chemicals.AbstractChemical = None
    """
    The chemical formula of the layer medium.
    If :obj:`None` (default), vacuum is assumed.
    """

    thickness: None | u.Quantity | na.AbstractScalar = None
    """The thickness of this layer"""

    interface: None | optika.materials.profiles.AbstractInterfaceProfile = None
    """
    The interface profile on the left side of this layer.
    """

    kwargs_plot: None | dict = None
    """Keyword arguments to be used in :meth:`plot` for styling of the layer."""

    x_label: None | u.Quantity = None
    """
    The horizontal coordinate of the label.
    If :obj:`None`, the label is plotted in the center of the layer
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.chemical),
            optika.shape(self.thickness),
            optika.shape(self.interface),
        )

    @property
    def _thickness_plot(self) -> u.Quantity | na.AbstractScalar:
        return self.thickness

    @property
    def _chemical(self) -> optika.chemicals.AbstractChemical:
        result = self.chemical
        if not isinstance(result, optika.chemicals.AbstractChemical):
            result = optika.chemicals.Chemical(result)
        return result

    def n(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> float | na.AbstractScalar:
        """
        The complex index of refraction of this entire layer.
        """
        if self.chemical is None:
            return 1
        else:
            return self._chemical.n(wavelength)

    @property
    def layer_sequence(self) -> LayerSequence:
        return LayerSequence([self])

    def transfer(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        polarized_s: bool | na.AbstractScalar,
        n: float | na.AbstractScalar,
        where: bool | na.AbstractScalar = True,
    ) -> tuple[
        na.AbstractScalar,
        na.AbstractScalar,
        na.Cartesian2dMatrixArray,
        na.AbstractScalar,
    ]:

        n_internal = self.n(wavelength)

        direction_internal = snells_law_scalar(
            cos_incidence=direction,
            index_refraction=n,
            index_refraction_new=n_internal,
        )

        refraction = matrices.refraction(
            wavelength=wavelength,
            direction_left=direction,
            direction_right=direction_internal,
            polarized_s=polarized_s,
            n_left=n,
            n_right=n_internal,
            interface=self.interface,
        )

        identity = na.Cartesian2dIdentityMatrixArray()
        refraction = np.where(where, refraction, identity)

        propagation = matrices.propagation(
            wavelength=wavelength,
            direction=direction_internal,
            thickness=self.thickness,
            n=n_internal,
        )

        where_propagation = np.abs(propagation.x.x) < 1e10
        where = where & where_propagation

        transfer = refraction @ propagation
        transfer = np.where(where, transfer, refraction)

        return n_internal, direction_internal, transfer, where

    def plot(
        self,
        x: u.Quantity = 0 * u.nm,
        z: u.Quantity = 0 * u.nm,
        width: u.Quantity = 1 * u.nm,
        ax: None | matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:
        """
        Plot this layer using  a :class:`matplotlib.patches.Rectangle` patch.

        Parameters
        ----------
        x
            The horizontal offset of the plotted layer
        z
            The vertical offset of the plotted layer.
        width
            The horizontal width of the plotted layer.
        ax
            The matplotlib axes instance on which to plot the layer.
        kwargs
            Additional keyword arguments to pass into
            :class:`matplotlib.patches.Rectangle`.
        """
        if ax is None:
            ax = plt.gca()

        chemical = self._chemical
        thickness = self.thickness
        kwargs_plot = self.kwargs_plot
        x_label = self.x_label

        if kwargs_plot is not None:
            kwargs = kwargs_plot | kwargs

        if x_label is None:
            x_label = x + width / 2
        else:
            x_label = x_label * width

        unit_x = na.unit(x)
        unit_z = na.unit(z)

        result = []
        if thickness is not None:

            patch = matplotlib.patches.Rectangle(
                xy=(x.to_value(unit_x), z.to_value(unit_z)),
                width=width.to_value(unit_x),
                height=thickness.to_value(unit_z),
                **kwargs,
            )

            result = ax.add_patch(patch)

            result = [result]

            if x_label <= x:
                ha = "right"
                x_arrow = 0 * unit_x
            elif x < x_label < x + width:
                ha = "center"
                x_arrow = x_label
            else:
                ha = "left"
                x_arrow = width

            y_arrow = y_label = z + thickness / 2

            formula = chemical.formula_latex

            ax.annotate(
                text=rf"{formula} (${thickness.value:0.0f}\,${thickness.unit:latex_inline})",
                xy=(x_arrow.to_value(unit_x), y_arrow.to_value(unit_z)),
                xytext=(x_label.to_value(unit_x), y_label.to_value(unit_z)),
                ha=ha,
                va="center",
                arrowprops=dict(
                    arrowstyle="->",
                ),
                transform=ax.get_yaxis_transform(),
            )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLayerSequence(
    AbstractLayer,
):
    """
    An interface describing a sequence of layers.
    """

    @property
    @abc.abstractmethod
    def layers(self) -> Sequence[AbstractLayer]:
        """A sequence of layers."""


@dataclasses.dataclass(eq=False, repr=False)
class LayerSequence(AbstractLayerSequence):
    """
    An explicit sequence of layers.

    Examples
    --------

    Plot a Si/Cr/Zr stack of layers

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import optika

        # Define the layer stack
        layers = optika.materials.LayerSequence([
            optika.materials.Layer(
                chemical="Si",
                thickness=10 * u.nm,
                kwargs_plot=dict(
                    color="tab:blue",
                    alpha=0.5,
                ),
            ),
            optika.materials.Layer(
                chemical="Cr",
                thickness=5 * u.nm,
                kwargs_plot=dict(
                    color="tab:orange",
                    alpha=0.5,
                ),
            ),
            optika.materials.Layer(
                chemical="Zr",
                thickness=15 * u.nm,
                kwargs_plot=dict(
                    color="tab:green",
                    alpha=0.5,
                ),
            ),
        ])

        # Plot the layer stack
        with astropy.visualization.quantity_support() as qs:
            fig, ax = plt.subplots(constrained_layout=True)
            layers.plot(ax=ax)
            ax.set_axis_off()
            ax.autoscale_view()
    """

    layers: Sequence[AbstractLayer] = dataclasses.MISSING
    """A sequence of layers."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            *[optika.shape(layer) for layer in self.layers],
        )

    def n(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        return self.layers[0].n(wavelength)

    @property
    def thickness(self) -> u.Quantity | na.AbstractScalar:
        result = 0 * u.nm
        for layer in self.layers:
            result = result + layer.thickness
        return result

    @property
    def _thickness_plot(self) -> u.Quantity | na.AbstractScalar:
        return self.thickness

    @property
    def interface(self) -> None | optika.materials.profiles.AbstractInterfaceProfile:
        return self.layers[0].interface

    def __getitem__(self, item: int | slice) -> LayerSequence:
        if isinstance(item, int):
            return self.layers[item]
        else:
            return type(self)(self.layers[item])

    @property
    def layer_sequence(self) -> LayerSequence:
        return self

    def transfer(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        polarized_s: bool | na.AbstractScalar,
        n: float | na.AbstractScalar,
        where: bool | na.AbstractScalar = True,
    ) -> tuple[
        na.AbstractScalar,
        na.AbstractScalar,
        na.Cartesian2dMatrixArray,
        na.AbstractScalar,
    ]:

        result = na.Cartesian2dIdentityMatrixArray()

        for layer in self.layers:
            n, direction, matrix_transfer, where = layer.transfer(
                wavelength=wavelength,
                direction=direction,
                polarized_s=polarized_s,
                n=n,
                where=where,
            )
            result = result @ matrix_transfer

        return n, direction, result, where

    def plot(
        self,
        x: u.Quantity = 0 * u.nm,
        z: u.Quantity = 0 * u.nm,
        width: u.Quantity = 10 * u.nm,
        ax: None | matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:

        z_current = z

        result = []
        for layer in reversed(self.layers):
            result += layer.plot(
                x=x,
                z=z_current,
                width=width,
                ax=ax,
                **kwargs,
            )
            z_current = z_current + layer._thickness_plot

        return result


@dataclasses.dataclass(eq=False, repr=False)
class PeriodicLayerSequence(AbstractLayerSequence):
    """
    A sequence of layers repeated an arbitrary number of times.

    This class is potentially much more efficient than :class:`LayerSequence`
    for large numbers of repeats.

    Examples
    --------

    Plot the periodic layer stack

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import optika

        # Define the layer stack
        layers = optika.materials.PeriodicLayerSequence(
            layers=[
                optika.materials.Layer(
                    chemical="Si",
                    thickness=10 * u.nm,
                    kwargs_plot=dict(
                        color="tab:blue",
                        alpha=0.5,
                    ),
                ),
                optika.materials.Layer(
                    chemical="Cr",
                    thickness=5 * u.nm,
                    kwargs_plot=dict(
                        color="tab:orange",
                        alpha=0.5,
                    ),
                ),
            ],
            num_periods=10,
        )

        # Plot the layer stack
        with astropy.visualization.quantity_support() as qs:
            fig, ax = plt.subplots(constrained_layout=True)
            layers.plot(ax=ax)
            ax.set_axis_off()
            ax.autoscale_view()
    """

    layers: Sequence[AbstractLayer] = dataclasses.MISSING
    """A sequence of layers."""

    num_periods: int = dataclasses.MISSING
    """The number of times to repeat the layer sequence."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            *[optika.shape(layer) for layer in self.layers],
        )

    def n(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        return self.layers[0].n(wavelength)

    @property
    def thickness(self) -> u.Quantity | na.AbstractScalar:
        return self.num_periods * LayerSequence(self.layers).thickness

    @property
    def _thickness_plot(self) -> u.Quantity | na.AbstractScalar:
        return LayerSequence(self.layers).thickness

    @property
    def interface(self) -> None | optika.materials.profiles.AbstractInterfaceProfile:
        return self.layers[0].interface

    @property
    def layer_sequence(self) -> LayerSequence:
        return LayerSequence(self.layers * self.num_periods)

    def transfer(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        polarized_s: bool | na.AbstractScalar,
        n: float | na.AbstractScalar,
        where: bool | na.AbstractScalar = True,
    ) -> tuple[
        na.AbstractScalar,
        na.AbstractScalar,
        na.Cartesian2dMatrixArray,
        na.AbstractScalar,
    ]:

        period = LayerSequence(self.layers)

        n, direction, start, where = period.transfer(
            wavelength=wavelength,
            direction=direction,
            polarized_s=polarized_s,
            n=n,
            where=where,
        )

        n, direction, periodic, where = period.transfer(
            wavelength=wavelength,
            direction=direction,
            polarized_s=polarized_s,
            n=n,
            where=where,
        )

        periodic = periodic.power(self.num_periods - 1)

        return n, direction, start @ periodic, where

    def plot(
        self,
        x: u.Quantity = 0 * u.nm,
        z: u.Quantity = 0 * u.nm,
        width: u.Quantity = 10 * u.nm,
        ax: None | matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:

        layers = LayerSequence(self.layers)

        result = layers.plot(
            x=x,
            z=z,
            width=width,
            ax=ax,
            **kwargs,
        )

        na.plt.brace_vertical(
            x=x,
            width=0.05 * width,
            beta=1 / (0.02 * layers.thickness),
            ymin=z,
            ymax=z + layers.thickness,
            ax=ax,
            label=rf"$\times {self.num_periods}$",
            kind="left",
            color="black",
        )

        return result



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika


@pytest.mark.parametrize(
    argnames="cos_incidence",
    argvalues=[
        1,
        na.linspace(-1, 1, axis="angle", num=7),
    ],
)
@pytest.mark.parametrize(
    argnames="index_refraction",
    argvalues=[
        1,
    ],
)
@pytest.mark.parametrize(
    argnames="index_refraction_new",
    argvalues=[
        1.5,
        1.5 + 0.2j,
        na.linspace(1, 2, axis="index_refraction_new", num=4),
    ],
)
def test_snells_law_scalar(
    cos_incidence: float | na.AbstractScalar,
    index_refraction: float | na.AbstractScalar,
    index_refraction_new: float | na.AbstractScalar,
):
    result = optika.materials.snells_law_scalar(
        cos_incidence=cos_incidence,
        index_refraction=index_refraction,
        index_refraction_new=index_refraction_new,
    )

    n1 = index_refraction
    n2 = index_refraction_new
    result_expected = np.cos(
        np.emath.arcsin(n1 * np.sin(np.emath.arccos(cos_incidence)) / n2)
    )

    assert np.allclose(result, result_expected)


@pytest.mark.parametrize(
    argnames="wavelength",
    argvalues=[
        350 * u.nm,
        na.linspace(300 * u.nm, 400 * u.nm, axis="wavelength", num=3),
    ],
)
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        na.Cartesian3dVectorArray(0, 0, 1),
    ],
)
@pytest.mark.parametrize(
    argnames="index_refraction",
    argvalues=[
        1,
    ],
)
@pytest.mark.parametrize(
    argnames="index_refraction_new",
    argvalues=[
        1.5,
        na.linspace(1, 2, axis="index_refraction_new", num=4),
    ],
)
@pytest.mark.parametrize(
    argnames="normal",
    argvalues=[
        None,
    ],
)
@pytest.mark.parametrize(
    argnames="is_mirror",
    argvalues=[
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    argnames="diffraction_order,spacing_rulings,normal_rulings",
    argvalues=[
        (
            0,
            None,
            None,
        ),
        (
            1,
            5 * u.um,
            na.Cartesian3dVectorArray(1, 0, 0),
        ),
        (
            na.linspace(-2, 2, "m", 5),
            5 * u.um,
            na.Cartesian3dVectorArray(1, 0, 0),
        ),
    ],
)
def test_snells_law(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    index_refraction: float | na.AbstractScalar,
    index_refraction_new: float | na.AbstractScalar,
    normal: None | na.AbstractCartesian3dVectorArray,
    is_mirror: bool | na.AbstractScalar,
    diffraction_order: int,
    spacing_rulings: None | u.Quantity | na.AbstractScalar,
    normal_rulings: None | na.AbstractCartesian3dVectorArray,
):
    result = optika.materials.snells_law(
        wavelength=wavelength,
        direction=direction,
        index_refraction=index_refraction,
        index_refraction_new=index_refraction_new,
        normal=normal,
        is_mirror=is_mirror,
        diffraction_order=diffraction_order,
        spacing_rulings=spacing_rulings,
        normal_rulings=normal_rulings,
    )
    if normal is None:
        normal = na.Cartesian3dVectorArray(0, 0, -1)

    assert isinstance(result, na.AbstractCartesian3dVectorArray)
    assert np.allclose(result.length, 1)
    if is_mirror:
        assert not np.allclose(np.sign(direction @ normal), np.sign(result @ normal))
    else:
        assert np.allclose(np.sign(direction @ normal), np.sign(result @ normal))



import abc
import dataclasses
import named_arrays as na
import optika
from . import AbstractLayer, AbstractMultilayerFilm, meshes

__all__ = [
    "AbstractThinFilmFilter",
    "ThinFilmFilter",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractThinFilmFilter(
    AbstractMultilayerFilm,
):
    """
    An interface describing a thin-film filter.
    """

    @property
    @abc.abstractmethod
    def layer(self) -> AbstractLayer:
        """The main layer of bulk material comprising this filter."""

    @property
    @abc.abstractmethod
    def layer_oxide(self) -> AbstractLayer:
        """The oxide layer that lies on either side of the bulk layer."""

    @property
    def layers(self) -> list[AbstractLayer]:
        layer = self.layer
        layer_oxide = self.layer_oxide
        return [
            layer_oxide,
            layer,
            layer_oxide,
        ]

    @property
    @abc.abstractmethod
    def mesh(self) -> meshes.AbstractMesh:
        """The mesh backing supporting this thin-film filter."""

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        result = super().efficiency(
            rays=rays,
            normal=normal,
        )
        return result * self.mesh.efficiency


@dataclasses.dataclass(eq=False, repr=False)
class ThinFilmFilter(
    AbstractThinFilmFilter,
):
    """
    A model of a thin-film EUV filter, such as those manufactured by Luxel
    :cite:p:`Powell1990`.

    Examples
    --------
    Plot the transmissivity of an aluminum thin-film filter, with a thickness
    of 100 nm and an oxide layer of 2 nm.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define a model of the thin-film filter
        film = optika.materials.ThinFilmFilter(
            layer=optika.materials.Layer(
                chemical="Al",
                thickness=100 * u.nm,
            ),
            layer_oxide=optika.materials.Layer(
                chemical="Al2O3",
                thickness=2 * u.nm,
            ),
            mesh=optika.materials.meshes.Mesh(
                chemical="Ni",
                efficiency=0.8,
                pitch=70 / u.mm,
            ),
        )

        # Define the wavelength of the incident light
        wavelength = na.linspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the rays incident on the filter
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(0, 0, 1),
        )

        # Compute the transmissivity of the filter
        transmissivity = film.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the transmissivity as a function of wavelength
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            transmissivity,
            ax=ax,
            axis="wavelength",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("transmissivity");
    """

    layer: AbstractLayer = dataclasses.MISSING
    """The main layer of bulk material comprising this filter."""

    layer_oxide: AbstractLayer = dataclasses.MISSING
    """The oxide layer that lies on either side of the bulk layer."""

    mesh: meshes.AbstractMesh = dataclasses.MISSING
    """The mesh backing supporting this thin-film filter."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.layer),
            optika.shape(self.layer_oxide),
            optika.shape(self.mesh),
        )



import abc
import dataclasses
import numpy as np
import scipy.special
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractInterfaceProfile",
    "ErfInterfaceProfile",
    "ExponentialInterfaceProfile",
    "LinearInterfaceProfile",
    "SinusoidalInterfaceProfile",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractInterfaceProfile(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """
    Abstract interface describing the :cite:t:`Stearns1989`
    interface profile between two layers in a multilayer stack.
    """

    @property
    @abc.abstractmethod
    def width(self) -> u.Quantity | na.AbstractScalar:
        """
        Characteristic length scale of the interface profile.
        """

    @abc.abstractmethod
    def __call__(self, z: u.Quantity | na.AbstractScalar) -> na.AbstractScalar:
        """
        Calculate the fraction of atoms that are in the new layer vs. those
        that are in the current layer.

        Parameters
        ----------
        z
            the depth in the current layer
        """

    @abc.abstractmethod
    def _derivative_fourier_transform(
        self,
        s: na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The Fourier transform of the derivative of the interface profile.
        This is used by :meth:`transmissivity` and :meth:`reflectivity` to
        compute the transmission and reflection coefficients of this
        interface profile.

        Parameters
        ----------
        s
            The effective wavenumber (independent variable of the Fourier
            transform).
        """

    def transmissivity(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction_before: float | na.AbstractScalar,
        direction_after: float | na.AbstractScalar,
        n_before: float | na.AbstractScalar,
        n_after: float | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The specular transmission amplitude for this interface profile.

        Parameters
        ----------
        wavelength
            The wavelength of the incident light in vacuum.
        direction_before
            The component of the incident light's propagation direction before
            the interface antiparallel to the surface normal.
        direction_after
            The component of the incident light's propagation direction after
            the interface antiparallel to the surface normal.
        n_before
            The complex index of refraction of the medium before the interface.
        n_after
            The complex index of refraction of the medium after the interface.

        Notes
        -----
        The specular transmission amplitude is given by :cite:t:`Stearns1989`
        Equation 42.
        """
        k_before = -2 * np.pi * n_before * direction_before / wavelength
        k_after = -2 * np.pi * n_after * direction_after / wavelength
        s = np.real(k_after - k_before)
        return self._derivative_fourier_transform(s)

    def reflectivity(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        n: float | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        Calculate the loss of the reflectivity due to this interface profile.

        Parameters
        ----------
        wavelength
            the wavelength of the incident light in vacuum
        direction
            The component of the incident light's propagation direction
            antiparallel to the surface normal.
        n
            The complex index of refraction of the medium before the interface.
        normal
            the vector perpendicular to the optical surface
        """
        k = -2 * np.pi * n * direction / wavelength
        s = np.real(-2 * k)
        return self._derivative_fourier_transform(s)


@dataclasses.dataclass(eq=False, repr=False)
class ErfInterfaceProfile(
    AbstractInterfaceProfile,
):
    r"""
    :cite:t:`Stearns1989` error function interface profile between two layers
    in a multilayer stack.

    The interface profile, :math:`p(z)` is defined as follows:

    .. math::

        p(z) = \frac{1}{\pi} \int_{-\infty}^z e^{-t^2/2 \sigma^2} dt

    Examples
    --------

    Plot an error function interface profile as a function of depth

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of widths
        width = na.linspace(1, 2, axis="width", num=5) * u.nm

        # Define the interface profile
        p = optika.materials.profiles.ErfInterfaceProfile(width=width)

        # Define an array of depths into the material
        z = na.linspace(-5, 5, axis="z", num=101) * u.nm

        # Plot the interface profile as a function of depth
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(z, p(z), ax=ax, axis="z", label=width);
        ax.set_xlabel(f"depth ({z.unit:latex_inline})");
        ax.set_ylabel(f"interface profile");
        ax.legend();

    Plot the reflectivity of the error function interface profile as a function
    of incidence angle

    .. jupyter-execute::

        # Define a wavelength
        wavelength = 304 * u.AA

        # Define an array of incidence angles
        angle = na.linspace(-90, 90, axis="angle", num=101) * u.deg

        # Define an array of direction cosines based off of the incidence angles
        direction = np.cos(angle)

        # Define the index of refraction of the current medium
        n = 1

        # Calculate the reflectivity for the given angles
        reflectivity = p.reflectivity(wavelength, direction, n)

        # Plot the reflectivity of the interface profile as a function of
        # incidence angle
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(angle, reflectivity, ax=ax, axis="angle", label=width);
        ax.set_xlabel(f"angle ({angle.unit:latex_inline})");
        ax.set_ylabel(f"reflectivity");
        ax.legend();

    """

    width: u.Quantity | na.AbstractScalar = 0 * u.nm
    r"""
    the width of the Gaussian in the intergrand of :math:`\text{erf}(x)`
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.width),
        )

    def __call__(self, z: u.Quantity | na.AbstractScalar) -> na.AbstractScalar:
        width = self.width
        x = z / (np.sqrt(2) * width)

        result = (1 + scipy.special.erf(x)) / 2

        return result

    def _derivative_fourier_transform(self, s: na.AbstractScalar):
        return np.exp(-np.square(s * self.width) / 2)


@dataclasses.dataclass(eq=False, repr=False)
class ExponentialInterfaceProfile(
    AbstractInterfaceProfile,
):
    r"""
    :cite:t:`Stearns1989` exponential function interface profile between two layers
    in a multilayer stack.

    The interface profile, :math:`p(z)` is defined as follows:

    .. math::

        p(z) = \begin{cases}
            \frac{1}{2} e^{\sqrt{2} z / \sigma}, & z \leq 0 \\
            1 - \frac{1}{2} e^{-\sqrt{2} z / \sigma}, & z \gt 0 \\
        \end{cases}

    Examples
    --------

    Plot an exponential interface profile as a function of depth

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of widths
        width = na.linspace(1, 2, axis="width", num=5) * u.nm

        # Define the interface profile
        p = optika.materials.profiles.ExponentialInterfaceProfile(width=width)

        # Define an array of depths into the material
        z = na.linspace(-5, 5, axis="z", num=101) * u.nm

        # Plot the interface profile as a function of depth
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(z, p(z), ax=ax, axis="z", label=width);
        ax.set_xlabel(f"depth ({z.unit:latex_inline})");
        ax.set_ylabel(f"interface profile");
        ax.legend();

    Plot the reflectivity of the exponential interface profile as a function
    of incidence angle

    .. jupyter-execute::

        # Define a wavelength
        wavelength = 304 * u.AA

        # Define an array of incidence angles
        angle = na.linspace(-90, 90, axis="angle", num=101) * u.deg

        # Define an array of direction cosines based off of the incidence angles
        direction = np.cos(angle)

        # Define the index of refraction of the current medium
        n = 1

        # Calculate the reflectivity for the given angles
        reflectivity = p.reflectivity(wavelength, direction, n)

        # Plot the reflectivity of the interface profile as a function of
        # incidence angle
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(angle, reflectivity, ax=ax, axis="angle", label=width);
        ax.set_xlabel(f"angle ({angle.unit:latex_inline})");
        ax.set_ylabel(f"reflectivity");
        ax.legend();

    """

    width: u.Quantity | na.AbstractScalar = 0 * u.nm
    r"""
    the width of the exponential
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.width),
        )

    def __call__(self, z: u.Quantity | na.AbstractScalar) -> na.AbstractScalar:
        width = self.width

        sgn_z = np.sign(z)
        result = (1 + sgn_z - sgn_z * np.exp(-sgn_z * np.sqrt(2) * z / width)) / 2

        return result

    def _derivative_fourier_transform(
        self,
        s: na.AbstractScalar,
    ) -> na.AbstractScalar:
        return 1 / (1 + np.square(s * self.width) / 2)


@dataclasses.dataclass(eq=False, repr=False)
class LinearInterfaceProfile(
    AbstractInterfaceProfile,
):
    r"""
    :cite:t:`Stearns1989` linear function interface profile between two layers
    in a multilayer stack.

    The interface profile, :math:`p(z)` is defined as follows:

    .. math::

        p(z) = \begin{cases}
            0, & z < -\sqrt{3} \sigma \\
            \frac{1}{2} + \frac{z}{2 \sqrt{3} \sigma}, & |z| \leq \sqrt{3} \sigma \\
            1, & z > \sqrt{3} \sigma
        \end{cases}

    Examples
    --------

    Plot an linear interface profile as a function of depth

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of widths
        width = na.linspace(1, 2, axis="width", num=5) * u.nm

        # Define the interface profile
        p = optika.materials.profiles.LinearInterfaceProfile(width=width)

        # Define an array of depths into the material
        z = na.linspace(-5, 5, axis="z", num=101) * u.nm

        # Plot the interface profile as a function of depth
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(z, p(z), ax=ax, axis="z", label=width);
        ax.set_xlabel(f"depth ({z.unit:latex_inline})");
        ax.set_ylabel(f"interface profile");
        ax.legend();

    Plot the reflectivity of the linear interface profile as a function
    of incidence angle

    .. jupyter-execute::

        # Define a wavelength
        wavelength = 304 * u.AA

        # Define an array of incidence angles
        angle = na.linspace(-90, 90, axis="angle", num=101) * u.deg

        # Define an array of direction cosines based off of the incidence angles
        direction = np.cos(angle)

        # Define the index of refraction of the current medium
        n = 1

        # Calculate the reflectivity for the given angles
        reflectivity = p.reflectivity(wavelength, direction, n)

        # Plot the reflectivity of the interface profile as a function of
        # incidence angle
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(angle, reflectivity, ax=ax, axis="angle", label=width);
        ax.set_xlabel(f"angle ({angle.unit:latex_inline})");
        ax.set_ylabel(f"reflectivity");
        ax.legend();

    """

    width: u.Quantity | na.AbstractScalar = 0 * u.nm
    """
    the width of the linear region
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.width),
        )

    def __call__(self, z: u.Quantity | na.AbstractScalar) -> na.AbstractScalar:
        width = self.width

        result = (1 / 2) + z / (2 * np.sqrt(3) * width)

        result = np.minimum(1, np.maximum(result, 0))

        return result

    def _derivative_fourier_transform(
        self,
        s: na.AbstractScalar,
    ) -> na.AbstractScalar:
        x = np.sqrt(3) * self.width * s
        result = np.sin(x.value) / x
        return result


@dataclasses.dataclass(eq=False, repr=False)
class SinusoidalInterfaceProfile(
    AbstractInterfaceProfile,
):
    r"""
    :cite:t:`Stearns1989` sinusoidal function interface profile between two layers
    in a multilayer stack.

    The interface profile, :math:`p(z)` is defined as follows:

    .. math::

        p(z) = \begin{cases}
            0, & z < -a \sigma \\
            \frac{1}{2} + \frac{1}{2} \sin \left( \frac{\pi z}{2 a \sigma} \right), & |z| \leq a \sigma \\
            1, & z > a \sigma
        \end{cases}

    where :math:`a = \pi / \sqrt{\pi^2 - 8}`.

    Examples
    --------

    Plot an sinusoidal interface profile as a function of depth

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of widths
        width = na.linspace(1, 2, axis="width", num=5) * u.nm

        # Define the interface profile
        p = optika.materials.profiles.SinusoidalInterfaceProfile(width=width)

        # Define an array of depths into the material
        z = na.linspace(-5, 5, axis="z", num=101) * u.nm

        # Plot the interface profile as a function of depth
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(z, p(z), ax=ax, axis="z", label=width);
        ax.set_xlabel(f"depth ({z.unit:latex_inline})");
        ax.set_ylabel(f"interface profile");
        ax.legend();

    Plot the reflectivity of the sinusoidal interface profile as a function
    of incidence angle

    .. jupyter-execute::

        # Define a wavelength
        wavelength = 304 * u.AA

        # Define an array of incidence angles
        angle = na.linspace(-90, 90, axis="angle", num=101) * u.deg

        # Define an array of direction cosines based off of the incidence angles
        direction = np.cos(angle)

        # Define the index of refraction of the current medium
        n = 1

        # Calculate the reflectivity for the given angles
        reflectivity = p.reflectivity(wavelength, direction, n)

        # Plot the reflectivity of the interface profile as a function of
        # incidence angle
        fig, ax = plt.subplots(constrained_layout=True);
        na.plt.plot(angle, reflectivity, ax=ax, axis="angle", label=width);
        ax.set_xlabel(f"angle ({angle.unit:latex_inline})");
        ax.set_ylabel(f"reflectivity");
        ax.legend();

    """

    width: u.Quantity | na.AbstractScalar = 0 * u.nm
    """
    the characteristic size of the sine wave
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.width),
        )

    def __call__(self, z: u.Quantity | na.AbstractScalar) -> na.AbstractScalar:
        width = self.width

        a = np.pi / (np.square(np.pi) - 8)
        z = np.minimum(a * width, np.maximum(z, -a * width))
        result = (1 / 2) + np.sin(np.pi * z / (2 * a * width) * u.rad) / 2

        return result

    def _derivative_fourier_transform(
        self,
        s: na.AbstractScalar,
    ) -> na.AbstractScalar:
        a = np.pi / (np.square(np.pi) - 8)
        x = a * self.width * s
        x1 = x - np.pi / 2
        x2 = x + np.pi / 2
        result = np.pi * (np.sin(x1 * u.rad) / x1 + np.sin(x2 * u.rad) / x2) / 4
        return result



from typing import TypeVar, Generic
import abc
import dataclasses
import numpy as np
import astropy.units as u  # type: ignore[import]
import named_arrays as na  # type: ignore[import]
import optika

__all__ = [
    "AbstractSag",
    "NoSag",
    "SphericalSag",
    "CylindricalSag",
    "AbstractConicSag",
    "ConicSag",
    "ParabolicSag",
    "ToroidalSag",
]

RadiusT = TypeVar(
    "RadiusT",
    bound=float | u.Quantity | na.AbstractScalar,
)
ConicT = TypeVar(
    "ConicT",
    bound=float | u.Quantity | na.AbstractScalar,
)
RadiusOfRotationT = TypeVar(
    "RadiusOfRotationT",
    bound=float | u.Quantity | na.AbstractScalar,
)
FocalLengthT = TypeVar(
    "FocalLengthT",
    bound=float | u.Quantity | na.AbstractScalar,
)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSag(
    optika.mixins.Printable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
):
    """
    Base class for all types of sag surfaces.
    """

    @property
    @abc.abstractmethod
    def parameters_slope_error(self) -> None | optika.metrology.SlopeErrorParameters:
        """collection of parameters to use when computing the slope error"""

    @property
    @abc.abstractmethod
    def parameters_roughness(self) -> None | optika.metrology.RoughnessParameters:
        """collection of parameters to use when computing the roughness"""

    @property
    @abc.abstractmethod
    def parameters_microroughness(self) -> None | optika.metrology.RoughnessParameters:
        """collection of parameters to use when computing the microroughness"""

    @abc.abstractmethod
    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        pass

    @abc.abstractmethod
    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        The vector perpendicular to the surface at the given position.

        Parameters
        ----------
        position
            The location on the surface to evaluate the normal vector
        """

    def intercept(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> optika.rays.AbstractRayVectorArray:
        """
        A set of new rays with the same direction as the input rays,
        but with the :attr:`optika.rays.RayVectorArray.position` updated to
        their interception point with this sag function.

        Parameters
        ----------
        rays
            input rays that will intercept this sag function
        """

        def line(t: na.AbstractScalar) -> na.Cartesian3dVectorArray:
            return rays.position + rays.direction * t

        def func(t: na.AbstractScalar) -> na.AbstractScalar:
            a = line(t)
            z = self(a)
            return a.z - z

        t_intercept = na.optimize.root_secant(
            function=func,
            guess=0 * u.mm,
        )

        result = rays.copy_shallow()
        result.position = line(t_intercept)
        return result


@dataclasses.dataclass(eq=False, repr=False)
class NoSag(
    AbstractSag,
):
    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    parameters_microroughness: None | optika.metrology.RoughnessParameters = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )

    @property
    def transformation(self) -> None:
        return None

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        return 0 * u.mm

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        return na.Cartesian3dVectorArray(0, 0, -1)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSphericalSag(
    AbstractSag,
    Generic[RadiusT],
):
    """
    Base class for all sag functions that can be approximated as a sphere.
    """

    @property
    @abc.abstractmethod
    def radius(self) -> RadiusT:
        """
        radius of curvature of the sag surface
        """

    @property
    def curvature(self) -> float | RadiusT:
        """
        The curvature of the spherical surface.
        Equal to the reciprocal of :attr:`radius`.
        """
        return 1 / self.radius


@dataclasses.dataclass(eq=False, repr=False)
class SphericalSag(
    AbstractSphericalSag,
    Generic[RadiusT],
):
    r"""
    A spherical sag function.

    The sag (:math:`z` coordinate) of a spherical surface is calculated using
    the expression

    .. math::

        z(x, y) = \frac{c (x^2 + y^2)}{1 + \sqrt{1 - c^2 (x^2 + y^2)}}

    where :math:`c` is the :attr:`curvature`,
    and :math:`x,y`, are the 2D components of the evaluation point.

    Examples
    --------
    Plot a slice through the sag surface

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        sag = optika.sags.SphericalSag(
            radius=na.linspace(100, 300, axis="radius", num=3) * u.mm,
        )

        position = na.Cartesian3dVectorArray(
            x=na.linspace(-90, 90, axis="x", num=101) * u.mm,
            y=0 * u.mm,
            z=0 * u.mm
        )

        z = sag(position)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            na.plt.plot(position.x, z, axis="x", label=sag.radius)
            plt.legend(title="radius")
    """

    radius: RadiusT = np.inf * u.mm
    transformation: None | na.transformations.AbstractTransformation = None
    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    parameters_microroughness: None | optika.metrology.RoughnessParameters = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.transformation),
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Evaluate the sag function at the given position

        Parameters
        ----------
        position
            point where the sag function will be calculated
        """
        c = self.curvature
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, position)
        c = na.broadcast_to(c, shape)
        position = na.broadcast_to(position, shape)

        r2 = np.square(position.x) + np.square(position.y)
        sz = c * r2 / (1 + np.sqrt(1 - np.square(c) * r2))
        return sz

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        c = self.curvature
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, position)
        c = na.broadcast_to(c, shape)
        position = na.broadcast_to(position, shape)

        x2, y2 = np.square(position.x), np.square(position.y)
        c2 = np.square(c)
        g = np.sqrt(1 - c2 * (x2 + y2))
        dzdx, dzdy = c * position.x / g, c * position.y / g
        result = na.Cartesian3dVectorArray(
            x=dzdx,
            y=dzdy,
            z=-1 * u.dimensionless_unscaled,
        )
        return result / result.length


@dataclasses.dataclass(eq=False, repr=False)
class CylindricalSag(
    AbstractSag,
    Generic[RadiusT],
):
    r"""
    A cylindrical sag function, where the local :math:`y` axis is the axis of
    symmetry for the cylinder.

    The sag (:math:`z` coordinate) of a spherical surface is calculated using
    the expression

    .. math::

        z(x, y) = \frac{c x^2}{1 + \sqrt{1 - c^2 x^2}}

    where :math:`c` is the :attr:`curvature`,
    and :math:`x` is the horizontal component of the evaluation point.

    Examples
    --------
    Plot a slice through the sag surface

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        sag = optika.sags.SphericalSag(
            radius=na.linspace(100, 300, axis="radius", num=3) * u.mm,
        )

        position = na.Cartesian3dVectorArray(
            x=na.linspace(-90, 90, axis="x", num=101) * u.mm,
            y=0 * u.mm,
            z=0 * u.mm
        )

        z = sag(position)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            na.plt.plot(position.x, z, axis="x", label=sag.radius)
            plt.legend(title="radius")
    """

    radius: RadiusT = np.inf * u.mm
    """The radius of the cylinder."""

    transformation: None | na.transformations.AbstractTransformation = None
    """
    The transformation between the surface coordinate system and the sag
    coordinate system.
    """

    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    """A set of parameters describing the slope error of the sag function."""

    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    """A set of parameters describing the roughness of the sag function."""

    parameters_microroughness: None | optika.metrology.RoughnessParameters = None
    """A set of parameters describing the microroughness of the sag function."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.transformation),
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:

        c = 1 / self.radius
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, position)
        c = na.broadcast_to(c, shape)
        position = na.broadcast_to(position, shape)

        r2 = np.square(position.x)
        sz = c * r2 / (1 + np.sqrt(1 - np.square(c) * r2))
        return sz

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:

        c = 1 / self.radius
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, position)
        c = na.broadcast_to(c, shape)
        position = na.broadcast_to(position, shape)

        x2 = np.square(position.x)
        c2 = np.square(c)
        g = np.sqrt(1 - c2 * x2)
        dzdx = c * position.x / g
        result = na.Cartesian3dVectorArray(
            x=dzdx,
            y=0,
            z=-1 * u.dimensionless_unscaled,
        )
        return result / result.length


@dataclasses.dataclass(eq=False, repr=False)
class AbstractConicSag(
    AbstractSag,
    Generic[RadiusT, ConicT],
):
    @property
    @abc.abstractmethod
    def radius(self) -> RadiusT:
        """the effective radius of the conic section"""

    @property
    @abc.abstractmethod
    def conic(self) -> ConicT:
        """conic constant of the conic section"""

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        c = 1 / self.radius
        conic = self.conic
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, conic, position)
        c = na.broadcast_to(c, shape)
        conic = na.broadcast_to(conic, shape)
        position = na.broadcast_to(position, shape)

        r2 = np.square(position.x) + np.square(position.y)
        sz = c * r2 / (1 + np.sqrt(1 - (1 + conic) * np.square(c) * r2))
        return sz

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        c = 1 / self.radius
        conic = self.conic
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(c, conic, position)
        c = na.broadcast_to(c, shape)
        conic = na.broadcast_to(conic, shape)
        position = na.broadcast_to(position, shape)

        x2 = np.square(position.x)
        y2 = np.square(position.y)
        c2 = np.square(c)
        g = np.sqrt(1 - (1 + conic) * c2 * (x2 + y2))
        dzdx, dzdy = c * position.x / g, c * position.y / g
        result = na.Cartesian3dVectorArray(
            x=dzdx,
            y=dzdy,
            z=-1 * u.dimensionless_unscaled,
        )
        return result / result.length


@dataclasses.dataclass(eq=False, repr=False)
class ConicSag(
    AbstractConicSag[RadiusT, ConicT],
):
    r"""
    Surface of revolution of a conic section

    The sag (:math:`z` coordinate) of a conic sag function is calculated using
    the expression

    .. math::

        z(x, y) = \frac{c (x^2 + y^2)}{1 + \sqrt{1 - c^2 (1 + k) (x^2 + y^2)}}

    where :math:`c` is the :attr:`curvature`,
    :math:`x,y`, are the 2D components of the evaluation point.
    and :math:`k` is the :attr:`conic` constant. See the table below for the
    meaning of the conic constant.

    ================== ==================
    conic constant     conic section type
    ================== ==================
    :math:`k < -1`     hyperbola
    :math:`k = -1`     parabola
    :math:`-1 < k < 0` ellipse
    :math:`k = 0`      sphere
    :math:`k > 0`      oblate ellipsoid
    ================== ==================

    Examples
    --------
    Plot a slice through the sag surface

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        sag = optika.sags.ConicSag(
            radius=100 * u.mm,
            conic=na.ScalarArray(
                ndarray=[-1.5, -1, -0.5, 0, 0.5] * u.dimensionless_unscaled,
                axes="conic",
            )
        )

        position = na.Cartesian3dVectorArray(
            x=na.linspace(-90, 90, axis="x", num=101) * u.mm,
            y=0 * u.mm,
            z=0 * u.mm
        )

        z = sag(position)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            na.plt.plot(position.x, z, axis="x", label=sag.conic)
            plt.legend(title="conic constant")
    """

    radius: RadiusT = np.inf * u.mm
    conic: ConicT = 0 * u.dimensionless_unscaled
    """the conic constant of the conic section"""
    transformation: None | na.transformations.AbstractTransformation = None
    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    parameters_microroughness: None | optika.metrology.RoughnessParameters = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.conic),
            optika.shape(self.transformation),
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )


@dataclasses.dataclass(eq=False, repr=False)
class ParabolicSag(
    AbstractConicSag[RadiusT, int],
):
    focal_length: FocalLengthT = np.inf * u.mm
    transformation: None | na.transformations.AbstractTransformation = None
    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    parameters_microroughness: None | optika.metrology.RoughnessParameters = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.focal_length),
            optika.shape(self.transformation),
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )

    @property
    def radius(self) -> RadiusT:
        return 2 * self.focal_length

    @property
    def conic(self) -> int:
        return -1


@dataclasses.dataclass(eq=False, repr=False)
class ToroidalSag(
    AbstractSphericalSag[RadiusT],
    Generic[RadiusT, RadiusOfRotationT],
):
    """
    A toroidal sag profile.
    """

    radius: RadiusT = np.inf * u.mm
    radius_of_rotation: RadiusOfRotationT = 0 * u.mm
    transformation: None | na.transformations.AbstractTransformation = None
    parameters_slope_error: None | optika.metrology.SlopeErrorParameters = None
    parameters_roughness: None | optika.metrology.RoughnessParameters = None
    parameters_microroughness: None | optika.metrology.RoughnessParameters = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.radius_of_rotation),
            optika.shape(self.transformation),
            optika.shape(self.parameters_slope_error),
            optika.shape(self.parameters_roughness),
            optika.shape(self.parameters_microroughness),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        c = self.curvature
        r = self.radius_of_rotation
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(position, c, r)
        position = na.broadcast_to(position, shape)
        c = na.broadcast_to(c, shape)
        r = na.broadcast_to(r, shape)

        x2 = np.square(position.x)
        y2 = np.square(position.y)
        zy = c * y2 / (1 + np.sqrt(1 - np.square(c) * y2))
        z = r - np.sqrt(np.square(r - zy) - x2)
        return z

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        c = self.curvature
        r = self.radius_of_rotation
        transformation = self.transformation
        if transformation is not None:
            position = transformation.inverse(position)

        shape = na.shape_broadcasted(position, c, r)
        position = na.broadcast_to(position, shape)
        c = na.broadcast_to(c, shape)
        r = na.broadcast_to(r, shape)

        x2 = np.square(position.x)
        y2 = np.square(position.y)
        c2 = np.square(c)
        g = np.sqrt(1 - c2 * y2)
        zy = c * y2 / (1 + g)
        f = np.sqrt(np.square(r - zy) - x2)
        dzdx = position.x / f
        dzydy = c * position.y / g
        dzdy = (r - zy) * dzydy / f
        result = na.Cartesian3dVectorArray(
            x=dzdx,
            y=dzdy,
            z=-1 * u.dimensionless_unscaled,
        )
        return result / result.length



"""
Meshes designed to support thin-film filters, such as the ones manufactured by
Luxel :cite:p:`Powell1990`.
"""

import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractMesh",
    "Mesh",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMesh(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """
    An interface describing the supporting mesh for a thin-film filter.
    """

    @property
    @abc.abstractmethod
    def chemical(self) -> str | optika.chemicals.AbstractChemical:
        """The chemical makeup of the mesh material."""

    @property
    @abc.abstractmethod
    def efficiency(self) -> float | na.AbstractScalar:
        """The fraction of light that is not blocked by the mesh."""

    @property
    @abc.abstractmethod
    def pitch(self) -> u.Quantity | na.AbstractScalar:
        """The density of the mesh in lines per inch or equivalent."""


@dataclasses.dataclass(eq=False, repr=False)
class Mesh(
    AbstractMesh,
):
    """
    An explicit description of a mesh designed to support a thin-film filter.
    """

    chemical: str | optika.chemicals.AbstractChemical = dataclasses.MISSING
    """The chemical formula of the mesh material."""

    efficiency: float | na.AbstractScalar = dataclasses.MISSING
    """The fraction of light that is not blocked by the mesh."""

    pitch: u.Quantity | na.AbstractScalar = dataclasses.MISSING
    """The density of the mesh in lines per inch or equivalent."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.chemical),
            optika.shape(self.efficiency),
            optika.shape(self.pitch),
        )



import pytest
import matplotlib.axes
import astropy.units as u
import named_arrays as na
import optika
from . import test_mixins
from . import test_propagators


surfaces = [
    optika.surfaces.Surface(),
    optika.surfaces.Surface(
        name="test_surface",
        sag=optika.sags.SphericalSag(radius=1000 * u.mm),
        material=optika.materials.Mirror(),
        aperture=optika.apertures.RectangularAperture(half_width=10 * u.mm),
        transformation=na.transformations.Cartesian3dTranslation(z=100 * u.mm),
    ),
    optika.surfaces.Surface(
        name="test_surface",
        sag=optika.sags.SphericalSag(radius=1000 * u.mm),
        material=optika.materials.Mirror(),
        aperture=optika.apertures.RectangularAperture(half_width=10 * u.mm),
        transformation=na.transformations.Cartesian3dTranslation(z=100 * u.mm),
        rulings=optika.rulings.Rulings(spacing=1 * u.um, diffraction_order=1),
    ),
]


class AbstractTestAbstractSurface(
    test_mixins.AbstractTestPlottable,
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
    test_mixins.AbstractTestShaped,
    test_propagators.AbstractTestAbstractLightPropagator,
):
    def test_name(self, a: optika.surfaces.AbstractSurface):
        if a.name is not None:
            assert isinstance(a.name, str)

    def test_sag(self, a: optika.surfaces.AbstractSurface):
        assert isinstance(a.sag, optika.sags.AbstractSag)

    def test_material(self, a: optika.surfaces.AbstractSurface):
        assert isinstance(a.material, optika.materials.AbstractMaterial)

    def test_aperture(self, a: optika.surfaces.AbstractSurface):
        if a.aperture is not None:
            assert isinstance(a.aperture, optika.apertures.AbstractAperture)

    def test_aperture_mechanical(self, a: optika.surfaces.AbstractSurface):
        if a.aperture_mechanical is not None:
            assert isinstance(a.aperture_mechanical, optika.apertures.AbstractAperture)

    def test_rulings(self, a: optika.surfaces.AbstractSurface):
        if a.rulings is not None:
            assert isinstance(a.rulings, optika.rulings.AbstractRulings)

    def test_is_field_stop(self, a: optika.surfaces.AbstractSurface):
        assert isinstance(a.is_field_stop, bool)

    def test_is_pupil_stop(self, a: optika.surfaces.AbstractSurface):
        assert isinstance(a.is_pupil_stop, bool)

    def test_is_stop(self, a: optika.surfaces.AbstractSurface):
        assert isinstance(a.is_stop, bool)

    class TestPlot(
        test_mixins.AbstractTestPlottable.TestPlot,
    ):
        def test_plot(
            self,
            a: optika.surfaces.AbstractSurface,
            ax: None | matplotlib.axes.Axes | na.ScalarArray,
            transformation: None | na.transformations.AbstractTransformation,
        ):
            if a.aperture is not None:
                if na.unit_normalized(a.aperture.wire()).is_equivalent(u.mm):
                    super().test_plot(
                        a=a,
                        ax=ax,
                        transformation=transformation,
                    )


@pytest.mark.parametrize("a", surfaces)
class TestSurface(
    AbstractTestAbstractSurface,
):
    pass



from __future__ import annotations
from typing import Sequence
import abc
import dataclasses
import named_arrays as na
import optika

__all__ = [
    "AbstractPropagator",
    "AbstractRayPropagator",
]


def propagate_rays(
    propagators: AbstractRayPropagator | Sequence[AbstractRayPropagator],
    rays: optika.rays.RayVectorArray,
) -> optika.rays.RayVectorArray:
    """
    Iterate through a sequence of ray propagators, calling
    :meth:`optika.propagators.AbstractRayPropagator.propagate_rays` on the given
    set of input rays.

    Parameters
    ----------
    propagators
        a sequence of ray propagators to interact with ``rays``
    rays
        the input rays to propagate through the sequence
    """
    if isinstance(propagators, AbstractRayPropagator):
        propagators = [propagators]

    for propagator in propagators:
        rays = propagator.propagate_rays(rays)

    return rays


def accumulate_rays(
    propagators: AbstractRayPropagator | Sequence[AbstractRayPropagator],
    rays: optika.rays.RayVectorArray,
    axis: str,
) -> optika.rays.RayVectorArray:
    """
    Iterate through a sequence of ray propagators, calling
    :meth:`optika.propagators.AbstractRayPropagator.propagate_rays` on the given
    set of input rays, and store the resulting the rays at every propagator.

    Parameters
    ----------
    propagators
        a sequence of ray propagators to interact with ``rays``
    rays
        the input rays to propagate through the sequence
    axis
        the new axis representing the sequence of propagators
    """
    if isinstance(propagators, AbstractRayPropagator):
        propagators = [propagators]

    result = []
    for propagator in propagators:
        rays = propagator.propagate_rays(rays)
        result.append(rays)

    result = na.stack(result, axis=axis)

    return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPropagator(
    abc.ABC,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRayPropagator(
    AbstractPropagator,
):
    @abc.abstractmethod
    def propagate_rays(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> optika.rays.AbstractRayVectorArray:
        """
        for the given input rays, calculate new rays based off of their
        interation with this object

        Parameters
        ----------
        rays
            a set of input rays that will interact with this object
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractLightPropagator(
    AbstractRayPropagator,
):
    pass



"""
Periodic rulings which can be added to an optical surface to model a
diffraction grating.
"""

from ._spacing import (
    AbstractRulingSpacing,
    ConstantRulingSpacing,
    Polynomial1dRulingSpacing,
    HolographicRulingSpacing,
)
from ._rulings import (
    AbstractRulings,
    Rulings,
    MeasuredRulings,
    SinusoidalRulings,
    SquareRulings,
    SawtoothRulings,
    TriangularRulings,
    RectangularRulings,
)

__all__ = [
    "AbstractRulingSpacing",
    "ConstantRulingSpacing",
    "Polynomial1dRulingSpacing",
    "HolographicRulingSpacing",
    "AbstractRulings",
    "Rulings",
    "MeasuredRulings",
    "SinusoidalRulings",
    "SquareRulings",
    "SawtoothRulings",
    "TriangularRulings",
    "RectangularRulings",
]



from __future__ import annotations
from typing import TypeVar, Generic
import abc
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractFieldVectorArray",
    "FieldVectorArray",
]

FieldT = TypeVar("FieldT", bound=na.AbstractCartesian2dVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractFieldVectorArray(
    na.AbstractCartesianVectorArray,
):
    @property
    @abc.abstractmethod
    def field(self) -> na.AbstractCartesian2dVectorArray:
        """
        the position of a point in the field of view
        """

    @property
    def type_abstract(self) -> type[AbstractFieldVectorArray]:
        return AbstractFieldVectorArray

    @property
    def type_explicit(self) -> type[FieldVectorArray]:
        return FieldVectorArray

    @property
    def type_matrix(self) -> type[na.PositionalMatrixArray]:
        raise NotImplementedError

    @property
    def explicit(self) -> FieldVectorArray:
        return super().explicit


@dataclasses.dataclass(eq=False, repr=False)
class FieldVectorArray(
    AbstractFieldVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[FieldT],
):
    field: FieldT = 0

    @classmethod
    def from_scalar(
        cls,
        scalar: na.ScalarLike,
        like: None | FieldVectorArray = None,
    ) -> FieldVectorArray:
        if like is None:
            type_like = cls
        else:
            type_like = type(like)
        return type_like(
            field=scalar,
        )
